# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'audio.audio_loader' module within the ketos library

    This module contains the utilities for loading waveforms and computing spectrograms.

    The audio representations currently implemented in Ketos are: 

    - :class:`Waveform <ketos.audio.waveform.Waveform>`
    - :class:`magnitude spectrogram <ketos.audio.spectrogram.MagSpectrogram>`
    - :class:`power spectrogram <ketos.audio.spectrogram.PowSpectrogram>`
    - :class:`mel spectrogram <ketos.audio.spectrogram.MelSpectrogram>`
    - :class:`CQT spectrogram <ketos.audio.spectrogram.CQTSpectrogram>`
    - :class:`CQTSpectrogram <ketos.audio.spectrogram.CQTSpectrogram>`
    - :class:`GammatoneFilterBank <ketos.audio.gammatone.GammatoneFilterBank>`
    - :class:`AuralFeatures <ketos.audio.gammatone.AuralFeatures>`
"""
import os
from pathlib import Path
import pandas as pd
import numpy as np
import soundfile as sf
import tarfile
import warnings
import shutil
from ketos.audio.waveform import Waveform, get_duration
from ketos.data_handling.data_handling import find_wave_files
from ketos.data_handling.selection_table import query
from ketos.utils import floor_round_up, ceil_round_down, user_format_warning


class ArchiveManager():
    ''' Class for extracting files from a .tar file.

        Use the method :meth:`ketos.audio.audio_loader.ArchiveManager.extract` to extract one or 
        several files from the .tar file to a temporary directory.

        Every time a file extraction request is submitted, the contents of the temporary directory 
        are updated as follows: 

         * Requested files *not* already present in the directory are extracted.
         * Requested files already present in the directory are left untouched.
         * Files present in the directory that are not part of the request are removed.

        At any given time, the location of the temporary directory and the paths of the files stored 
        within the directory can be accessed via the attributes @extract_dir and @extracted_files.

        Args:
            tar_path: str
                Path to the .tar file
            extract_dir: str
                Path to the directory where the extracted files are temporarily stored. The directory 
                is automatically created. If a directory already exists at the specified path, all its 
                contents will be deleted. By default, audio files are extracted to the folder `kt-tmp` 
                within the current working directory.

        Attributes:
            tar: TarFile
                tar object
            tar_path: str
                Path to the .tar file
            extract_dir: str
                Path to the directory where the extracted files are temporarily stored
            extracted_files: list
                Relative paths to the currently extracted files
    '''
    def __init__(self, tar_path, extract_dir="kt-tmp"):
        self.tar_path = tar_path
        self.tar = tarfile.open(tar_path)
        self.extract_dir = extract_dir
        self.extracted_files = []
        self.close() 

    def _extract_files(self, paths):
        """ Helper function for extracting files.

            Issues a UserWarning if a file does not exist at the specified 
            path within the tar archive.

            Args:
                paths: list
                    Relative paths to the files to be extracted from within the tar archive
        """ 
        for path in paths:
            try:
                self.tar.extract(member=path, path=self.extract_dir)
                self.extracted_files.append(path)
            except KeyError as e:
                warnings.formatwarning = user_format_warning
                warnings.warn(f"{path} not found in {self.tar_path}")

    def _remove_files(self, paths):
        """ Helper function for removing files from the extraction directory.

            Args:
                paths: list
                    Relative paths to the files to be removed
        """ 
        for path in paths:
            dst = os.path.join(self.extract_dir, path)
            os.remove(dst)
            self.extracted_files.remove(path)

    def extract(self, paths):
        """ Update the files in the extraction directory.

            Every time this method is called, the contents of the temporary directory 
            are updated as follows: 

                * Requested files *not* already present in the directory are extracted.
                * Requested files already present in the directory are left untouched.
                * Files present in the directory that are not part of the request are removed.

            Args:
                paths: str or list
                    Relative path(s) of the files within the tar archive that we want to 
                    be available in the extraction directory
        """
        if isinstance(paths, str):
            paths = [paths]

        paths_extract = [path for path in paths if path not in self.extracted_files]
        paths_remove = [path for path in self.extracted_files if path not in paths]
        self._extract_files(paths_extract)
        self._remove_files(paths_remove)

    def close(self):
        """ Remove the extraction directory and its contents 
        """
        if os.path.exists(self.extract_dir):
            shutil.rmtree(self.extract_dir)

class SelectionGenerator():
    """ Template class for selection generators.
    """
    def __iter__(self):
        return self

    def __next__(self):
        """ Returns the next audio selection.

            An audio selection is represented as a dictionary 
            with two required keys (data_dir, filename) and 
            an unlimited number of optional keys, which typically 
            include offset, duration, and label.
        
            Must be implemented in child class.

            Returns:
                : dict()
                    Next audio selection
        """
        pass

    def num(self):
        """ Returns total number of selections.
        
            Must be implemented in child class.

            Returns:
                : int
                    Total number of selections.
        """
        pass

    def reset(self):
        """ Resets the selection generator to the beginning.
        """        
        pass
    

class SelectionTableIterator(SelectionGenerator):
    """ Iterates over entries in a selection table.

        Args: 
            data_dir: str
                Path to top folder containing audio files, or a .tar archive file.
            selection_table: pandas DataFrame
                Selection table
            include_attrs: bool
                If True, load data from all attribute columns in the selection table. Default is False.
            attrs: list(str)
                Specify the names of the attribute columns that you wish to load data from. 
                Overwrites include_attrs if specified. If None, all columns will be loaded provided that 
                include_attrs=True.
            extract_dir: str
                Temporary directory for storing audio files extracted from a tar archive file. 
                Only relevant if @data_dir points to a .tar file. The directory will be automatically 
                created. If a directory already exists at the specified path, all its contents will be 
                deleted. By default, audio files are extracted to the folder `kt-tmp` within the current
                working directory. Note that this folder must be deleted manually when it is no longer needed.
    """
    def __init__(self, data_dir, selection_table, include_attrs=False, attrs=None, extract_dir="kt-tmp"):
        self.sel = selection_table

        if os.path.isfile(data_dir) and tarfile.is_tarfile(data_dir):
            self.tar = ArchiveManager(data_dir, extract_dir)
            self.dir = self.tar.extract_dir
        else:
            self.tar = None
            self.dir = data_dir

        self.counter = 0

        all_attrs = list(self.sel.columns.values)
        for col in ['start', 'end', 'label']: 
            if col in all_attrs: all_attrs.remove(col)

        if attrs is not None:
            for col in attrs: 
                if col not in all_attrs: attrs.remove(col)
            self.attrs = attrs
        elif include_attrs:
            self.attrs = all_attrs
        else:
            self.attrs = []

        # determine if the selection table has been formatted according to 
        # the new ketos style (>=2.6.0) or the old style 
        self._new_style = (self.sel.index.names[0] == "sel_id")

        if self._new_style:
            self.sel_ids = self.sel.index.get_level_values(0).unique()
            self.num_sel = len(self.sel_ids)
        else:
            self.num_sel = len(self.sel)

    def __next__(self):
        """ Returns the next audio selection.

            Returns:
                audio_sel: dict
                    Audio selection
        """
        audio_sel = self.get_selection(self.counter)

        if self.tar is not None:
            self.tar.extract(audio_sel['filename'])

        self.counter = (self.counter + 1) % self.num() #update selection counter
        return audio_sel

    def num(self):
        """ Returns total number of selections.
        
            Returns:
                : int
                    Total number of selections.
        """
        return self.num_sel

    def reset(self):
        """ Resets the selection generator to the beginning of the selection table.
        """        
        self.counter = 0
        if self.tar is not None:
            self.tar.close()
        
    def get_selection(self, n):
        """ Returns the n-th audio selection in the table.

            Args:
                n: int
                    The index (0,1,2,...) of the desired selection.

            Returns:
                res: dict
                    The selection
        """
        res = {'data_dir': self.dir}

        if self._new_style:
            selection = self.sel.loc[self.sel_ids[n]]
            res['filename'] = selection.index.values            
        
        else:
            selection = self.sel.iloc[n]
            res['filename'] = self.sel.index.values[n][0]
        
        # start time
        if 'start' in selection.keys(): 
            offset = selection['start']
        else: 
            offset = 0

        # duration
        if 'end' in selection.keys(): 
            duration = selection['end'] - offset
        else:
            duration = None

        # pass offset and duration to return dict
        res['offset'] = offset
        if duration is not None:
            res['duration'] = duration

        # label
        if 'label' in self.sel.columns.values: 
            res['label'] = selection['label']

        # attribute columns
        for col in self.attrs: 
            res[col] = selection[col]

        # for new style, convert pandas Series to numpy arrays
        if self._new_style:
            for key in res.keys():
                if isinstance(res[key], pd.Series):
                    res[key] = res[key].values
                    if key in ["offset", "duration"]:
                        res[key] = res[key].astype(float) #ensure float

        # OBS: for labels and attributes, use only the first entry
        for col in ["label"] + self.attrs:
            if col in res.keys() and np.ndim(res[col]) > 0:
                res[col] = res[col][0]
        
        return res


class FrameStepper(SelectionGenerator):
    """ Generates selections with uniform length 'duration', with successive selections 
        displaced by a fixed amount 'step' (If 'step' is not specified, it is set equal 
        to 'duration'.)

        Args: 
            duration: float
                Selection length in seconds.
            step: float
                Separation between consecutive selections in seconds. If None, the step size 
                equals the selection length.
            path: str
                Path to folder containing .wav files. If None is specified, the current directory will be used.
            filename: str or list(str)
                Relative path to a single .wav file or a list of .wav files. Optional.
            pad: bool
                If True (default), the last segment is allowed to extend beyond the endpoint of the audio file.
    """
    def __init__(self, duration, step=None, path=None, filename=None, pad=True):            
        self.duration = duration
        if step is None: self.step = duration
        else: self.step = step

        if path is None: path = os.getcwd()

        # get all wav files in the folder, including subfolders
        if filename is None:
            self.dir = path
            self.files = find_wave_files(path=path, return_path=True, search_subdirs=True)
            assert len(self.files) > 0, '{0} did not find any wave files in {1}'.format(self.__class__.__name__, path)

        else:
            if isinstance(filename, str):
                fullpath = os.path.join(path,filename)
                assert os.path.exists(fullpath), '{0} could not find {1}'.format(self.__class__.__name__, fullpath)
                self.dir = os.path.dirname(fullpath)
                self.files = [os.path.basename(fullpath)]
            else:                
                assert isinstance(filename, list), 'filename must be str or list(str)'        
                self.dir = path
                self.files = filename

        # get file durations
        self.file_durations = np.array(get_duration([os.path.join(self.dir, f) for f in self.files]))

        # discard any files with 0 second duration
        # self.files = np.array(self.files)[self.file_durations > 0].tolist()
        # self.file_durations = self.file_durations[self.file_durations > 0]

        # obtain file durations and compute number of frames for each file
        self.num_segs = np.maximum((self.file_durations - self.duration) / self.step + 1, 1)
        if pad:
            # self.num_segs = ceil_round_down(self.num_segs, decimals=6).astype(int)   
            self.num_segs = np.ceil(np.around(self.num_segs, decimals=6)).astype(int)     
        else:
            # self.num_segs = floor_round_up(self.num_segs, decimals=6).astype(int)     
            self.num_segs = np.floor(np.around(self.num_segs, decimals=6)).astype(int)

        self.num_segs_tot = np.sum(self.num_segs)

        self.reset()

    def __next__(self):
        """ Returns the next audio selection.
        
            Returns:
                audio_sel: dict
                    Audio selection
        """
        audio_sel = {'data_dir':self.dir, 'filename': self.files[self.file_id], 'offset':self.time, 'duration':self.duration}
        self.time += self.step #increment time       
        self.seg_id += 1 #increment segment ID
        if self.seg_id == self.num_segs[self.file_id]: self._next_file() #if this was the last segment, jump to the next file
        return audio_sel

    def num(self):
        """ Returns total number of selections.
        
            Returns:
                : int
                    Total number of selections.
        """
        return self.num_segs_tot

    def _next_file(self):
        """ Jump to next file. 
        """
        self.file_id = (self.file_id + 1) % len(self.files) #increment file ID
        self.seg_id = 0 #reset
        self.time = 0 #reset

    def reset(self):
        """ Resets the selection generator to the beginning of the first file.
        """        
        self.file_id = -1
        self._next_file()

    def get_file_paths(self, fullpath=True):
        """ Get the paths to the audio files associated with this instance.

            Args:
                fullpath: bool
                    Whether to return the full path (default) or only the filename.

            Returns:
                ans: list
                    List of file paths
        """
        if fullpath:
            ans = [os.path.join(self.dir, f) for f in self.files]
        else:
            ans = self.files

        return ans

    def get_file_durations(self):
        """ Get the durations of the audio files associated with this instance.

            Returns:
                ans: list
                    List of file durations in seconds
        """
        return self.file_durations.tolist()


def _file_limits_warning(start, end, file_path, file_duration):
    """ Helper function for the AudioLoader class.
    
        Generates warning messages if the selection start or end time is outside of the audio file's limits.

        Args:
            start: float
                Selection start time with respect to beginning of file in seconds.
            end: float
                Selection end time with respect to beginning of file in seconds.
            file_path: str
                Full path of the audio file.
            file_duration: float
                Audio file length in seconds.
    """
    # total length of selection
    if file_duration == 0:
        return
    len_tot = end - start
    # determine how much of the selection is outside the file
    len_outside = max(0, -start) + max(0, end - file_duration)

    warnings.formatwarning = user_format_warning

    # print warnings if selection end is zero or negative
    file_info = f"While processing {os.path.basename(file_path)}"
    if (end <= 0):
        warnings.warn(f"{file_info}: selection has negative end time ({end:.2f}s).")

    # print warnings if selection start is later than the file end time
    elif (start > file_duration):
        warnings.warn(f"{file_info}: selection start time exceeds file duration ({start:.2f}s).")

    #print a warning that the selection has 0 or negative length (end before start)
    elif (len_tot <= 0):
        warnings.warn(f"{file_info}: selection has negative duration ({start:.2f},{end:.2f}).")
         
    # print a warning that a fraction larger than 50% of the selection is outside the file
    elif (len_outside > 0.5 * len_tot):
        warnings.warn(f"{file_info}: over 50% of the selection falls outside the audio file ({start:.2f}s,{end:.2f}s).")


class AudioLoader():
    """ Class for loading segments of audio data from .wav files. 

        Several representations of the audio data are possible, including 
        waveform, magnitude spectrogram, power spectrogram, mel spectrogram, 
        and CQT spectrogram.

        Args:
            selection_gen: SelectionGenerator
                Selection generator
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            annotations: pandas DataFrame
                Annotation table
            representation: class or list of classes
                Audio data representation. This is a class that must receive the raw audio data and will transform the data 
                into the specified audio representation object.
                
                Classes available in ketos:

                    * Waveform: 
                        (rate), (resample_method)
                    
                    * MagSpectrogram, PowerSpectrogram, MelSpectrogram: 
                        audio, window, step, (window_func), (rate), (resample_method)
                    
                    * CQTSpectrogram:
                        audio, step, bins_per_oct, (freq_min), (freq_max), (window_func), (rate), (resample_method)

                It is also possible to specify multiple audio presentations as a list.
            representation_params: dict or list of dict
                Dictionary containing any required and optional arguments for the representation class. If more than one
                representation is given `representation_params` must be a list of the same length and in the same order.
            batch_size: int
                Load segments in batches rather than one at the time. 
            stop: bool
                Raise StopIteration when all selections have been loaded. Default is True.
        
        Examples:

            Creating an AudioLoader to load selections:         
            
            >>> from ketos.audio.audio_loader import AudioLoader, SelectionTableIterator
            >>> from ketos.data_handling.selection_table import use_multi_indexing
            >>> import pandas as pd
            >>> # Load the audio representation you want to pass
            >>> from ketos.audio.spectrogram import MagSpectrogram
            >>> # specify the audio representation
            >>> rep = {'window':0.2, 'step':0.02, 'window_func':'hamming'}
            >>> # Load selections
            >>> sel = pd.DataFrame({'filename':["2min.wav", "2min.wav"],'start':[0.10,0.12],'end':[0.46,0.42]})
            >>> sel = use_multi_indexing(sel, 'sel_id')
            >>> # create a generator for iterating over all the selections 
            >>> generator = SelectionTableIterator(data_dir="ketos/tests/assets/", selection_table=sel)
            >>> # Create a loader by passing the generator and the representation to the AudioLoader
            >>> loader = AudioLoader(selection_gen=generator, representation=MagSpectrogram, representation_params=rep)
            >>> # print number of segments
            >>> print(loader.num())
            2
            >>> # load and plot the first selection
            >>> spec = next(loader)
            >>>
            >>> import matplotlib.pyplot as plt
            >>> fig = spec.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/spec_loader_2min_0.png")
            >>> plt.close(fig)
            
            .. image:: ../../../ketos/tests/assets/tmp/spec_loader_2min_0.png

            Creating an AudioLoader to load selections made from annotations:   

            >>> from ketos.audio.audio_loader import AudioLoader, SelectionTableIterator
            >>> from ketos.data_handling.selection_table import standardize
            >>> import pandas as pd
            >>> # Load the audio representation you want to pass
            >>> from ketos.audio.spectrogram import MagSpectrogram
            >>> # specify the audio representation
            >>> rep = {'window':0.2, 'step':0.02, 'window_func':'hamming'}
            >>> # Load selections
            >>> annot = pd.DataFrame([{"filename":"2min.wav", "start":2.0, "end":3.0, "label":0},
            ...         {"filename":"2min.wav", "start":5.0, "end":6.0, "label":0},
            ...         {"filename":"2min.wav", "start":21.0, "end":22.0, "label":0},
            ...         {"filename":"2min.wav", "start":25.0, "end":27.0, "label":0}])
            >>> annot_std = standardize(table=annot)
            >>> # create a generator for iterating over all the selections 
            >>> generator = SelectionTableIterator(data_dir="ketos/tests/assets/", selection_table=annot_std)
            >>> # Create a loader by passing the generator and the representation to the AudioLoader
            >>> loader = AudioLoader(selection_gen=generator, representation=MagSpectrogram, representation_params=rep)
            >>> # print number of segments
            >>> print(loader.num())
            4
            >>> # load and plot the first selection
            >>> spec = next(loader)
            >>>
            >>> import matplotlib.pyplot as plt
            >>> fig = spec.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/spec_loader_2min_1.png")
            >>> plt.close(fig)
            
            .. image:: ../../../ketos/tests/assets/tmp/spec_loader_2min_1.png

            For more examples see child class :class:`audio.audio_loader.AudioFrameLoader`   

    """
    def __init__(self, selection_gen, channel=0, annotations=None, representation=Waveform, representation_params=None, 
                        batch_size=1, stop=True, **kwargs):

        self.representation = representation
        self.representation_params = representation_params

        if not isinstance(self.representation, list):
            self.representation = [self.representation]
            self.representation_params = [self.representation_params]
        
        for i in range(len(self.representation)):
            if self.representation_params[i] == None: # If no parameters are given then create an empty dict (this will use the default params)
                self.representation_params[i] = {}
        self.channel = channel
        self.selection_gen = selection_gen
        self.annot = annotations
        # QUESTION: kwargs is carrying more optional arguments. such as compute phase... it feels very wrong. shouldnt the phase be another representation? or an arugment of the spectrogram class?
        self.kwargs = kwargs
        self.batch_size = batch_size
        self.stop = stop
        self.file_durations = dict()
        self.reset()

    def __iter__(self):
        return self

    def __next__(self):
        """ Load next audio segment or batch of audio segments.

            Depending on how the loader was initialized, the return value can either be 
            an instance of :class:`BaseAudio <ketos.audio.base_audio.BaseAudio>` (or, 
            more commonly, a instance of one of its derived classes such as the 
            :class:`Waveform <ketos.audio.waveform.Waveform>` or 
            :class:`MagSpectrogram <ketos.audio.spectrogram.MagSpectrogram>`
            classes), a list of such objects, or a nested listed of such objects. 

            Some examples:

             * If the loader was initialized with the audio representation `representation=Waveform`,   
               `representation_params=None` (default) and with `batch_size=1` (default), the return  
               value will be a single instance of :class:`Waveform <ketos.audio.waveform.Waveform>`.

             * If the loader was initialized with the audio representation 
               `representation=[Waveform, MagSpectrogram]`, `representation_params=[None, {'window':0.1,'step':0.02}]` 
               and with `batch_size=1` (default), the return value will be a list 
               of length 2, where the first entry holds an instance of 
               :class:`Waveform <ketos.audio.waveform.Waveform>` and the second entry holds an instance 
               of :class:`MagSpectrogram <ketos.audio.spectrogram.MagSpectrogram>`.

             * If the loader was initialized with the audio representation 
               `representation=[Waveform, MagSpectrogram]`, `representation_params=[None, {'window':0.1,'step':0.02}]` 
               and with `batch_size>1`, the return value will be a nested list with outer 
               length equal to `batch_size` and inner length 2, corresponding to the number of 
               audio representations.

            If the loader was initialized with `stop=True` this method will raise `StopIteration` 
            when all the selections have been loaded.

            Returns: 
                a: BaseAudio, list(BaseAudio), or list(list(BaseAudio))
                    Next segment or next batch of segments
        """
        return self._next_batch(load=True)

            
    def skip(self):
        """ Skip to the next audio segment or batch of audio segments
            without loading the current one.
        """        
        self._next_batch(load=False)

    def _next_batch(self, load=True):
        """ Load next audio segment or batch of audio segments.

            Helper function for :meth:`__next()__` and :meth:`skip()`.

            Args:
                load: bool
                    Whether to load the audio data.
        """
        if self.counter == self.num():
            if self.stop:
                raise StopIteration
            else:
                self.reset()

        a = []
        for _ in range(self.batch_size):
            if self.counter < self.num():
                selection = next(self.selection_gen)
                if load:
                    a.append(self.load(**selection, **self.kwargs))
                self.counter += 1

        if load:
            if self.batch_size == 1: a = a[0]
            return a

    def num(self):
        """ Returns total number of segments.
        
            Returns:
                : int
                    Total number of segments.
        """
        return self.selection_gen.num()

    def load(self, data_dir, filename, offset=0, duration=None, label=None, **kwargs):
        """ Load audio segment for specified file and time.

            Args:
                data_dir: str
                    Data directory
                filename: str
                    Filename or relative path
                offset: float
                    Start time of the segment in seconds, measured from the 
                    beginning of the file.
                duration: float
                    Duration of segment in seconds.
                label: int
                    Integer label
        
            Returns: 
                seg: BaseAudio or list(BaseAudio)
                    Audio segment
        """
        # convert scalar args to arrays
        if np.ndim(filename) == 0:
            filename = [filename]
            offset = np.array([offset], dtype=float)
            if duration is None:
                duration = [None]
            else:
                duration = np.array([duration], dtype=float)

        path = [str(Path(data_dir, fname).resolve()) for fname in filename]
        id = filename[0]

        # issue warnings if selections extend beyond file limits
        for i in range(len(path)):  
            p = path[i]          
            file_duration = self.file_durations.get(p)
            # If file duration does not exist in the dict, get file duration and add it to the dict
            if file_duration == None:
                file_duration = get_duration(p)[0]
                self.file_durations[p] = file_duration

            start = offset[i]
            end = file_duration - start if duration[i] is None else start + duration[i]
            _file_limits_warning(start=start, end=end, file_path=p, file_duration=file_duration)


        # load audio
        segs = []
        for i in range(len(self.representation)): # For each representation
            self.representation_params[i]['duration'] = duration # The duration for the representation is defined by each segment
            seg = self.representation[i].from_wav(path=path, channel=self.channel, offset=offset, id=id, **self.representation_params[i], **kwargs) 

            # add label
            if label is not None:
                seg.label = label

            # add annotations
            if self.annot is not None:
                file_offset = np.concatenate([[0],np.cumsum(duration)[:-1]])
                for j in range(len(filename)):
                    q = query(self.annot, filename=filename[j], start=offset[j], end=offset[j]+duration[j])
                    if len(q) > 0:
                        q['start'] += file_offset[j] - offset[j]
                        q['end']   += file_offset[j] - offset[j]
                        seg.annotate(df=q)  
            
            # We can add the metadata information such as start, duration and filename of spectrogram directly to the object here.
            # The problem is the duration wich may be different than what the user set depending if the representation needs to add some extra seconds
            # filename is just the filename and not the full path... not sure if we should change this? what are the advantages of giving the filename over path
            seg.start = offset[0]
            seg.filename = id
            segs.append(seg)

        if len(segs) == 1: segs = segs[0]

        return segs

    def reset(self):
        """ Resets the audio loader to the beginning.
        """        
        self.selection_gen.reset()
        self.counter = 0


class AudioFrameLoader(AudioLoader):
    """ Load audio segments by sliding a fixed-size frame across the recording.

        The frame size is specified with the 'duration' argument, while the 'step'
        argument may be used to specify the step size. (If 'step' is not specified, 
        it is set equal to 'duration'.)

        Args:
            duration: float
                Segment duration in seconds. 
            step: float
                Separation between consecutive segments in seconds. If None, the step size 
                equals the segment duration. 
            path: str
                Path to folder containing .wav files. If None is specified, the current directory will be used.
            filename: str or list(str)
                relative path to a single .wav file or a list of .wav files. Optional
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            annotations: pandas DataFrame
                Annotation table
            representation: class or list of classes
                Audio data representation. This is a class that must receive the raw audio data 
                and will transform the data into the specified audio representation object.  
                It is also possible to specify multiple audio presentations as a list. These 
                presentations must have the same duration.
            representation_params: dict or list of dict
                Dictionary containing any required and optional arguments for the representation class. If more than one
                representation is given `representation_params` must be a list of the same length and in the same order.
            batch_size: int
                Load segments in batches rather than one at the time. 
            stop: bool
                Raise StopIteration if the iteration exceeds the number of available selections. Default is False.
            pad: bool
                If True (default), the last segment is allowed to extend beyond the endpoint of the audio file.

        Examples:
            >>> from ketos.audio.audio_loader import AudioFrameLoader
            >>> # Load the audio representation you want to pass
            >>> from ketos.audio.spectrogram import MagSpectrogram
            >>> # specify path to wav file
            >>> filename = 'ketos/tests/assets/2min.wav'
            >>> # check the duration of the audio file
            >>> from ketos.audio.waveform import get_duration
            >>> print(get_duration(filename)[0])
            120.832
            >>> # specify the audio representation parameters
            >>> rep = {'window':0.2, 'step':0.02, 'window_func':'hamming', 'freq_max':1000.}
            >>> # create an object for loading 30-s long spectrogram segments, using a step size of 15 s (50% overlap) 
            >>> loader = AudioFrameLoader(duration=30., step=15., filename=filename, representation=MagSpectrogram, representation_params=rep)
            >>> # print number of segments
            >>> print(loader.num())
            8
            >>> # load and plot the first segment
            >>> spec = next(loader)
            >>>
            >>> import matplotlib.pyplot as plt
            >>> fig = spec.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/spec_2min_0.png")
            >>> plt.close(fig)
            
            .. image:: ../../../ketos/tests/assets/tmp/spec_2min_0.png
    """
    def __init__(self, duration, step=None, path=None, filename=None, channel=0, 
                    annotations=None, representation=Waveform, representation_params=None, batch_size=1, 
                    stop=True, pad=True):

        if batch_size > 1:
            print("Warning: batch_size > 1 results in different behaviour for ketos versions >= 2.4.2 than earlier \
                   versions. You may want to check out the AudioFrameEfficientLoader class.")

        super().__init__(selection_gen=FrameStepper(duration=duration, step=step, path=path, pad=pad, filename=filename), 
            channel=channel, annotations=annotations, representation=representation, representation_params=representation_params, 
            batch_size=batch_size, stop=stop)

    def get_file_paths(self, fullpath=True):
        """ Get the paths to the audio files associated with this instance.

            Args:
                fullpath: bool
                    Whether to return the full path (default) or only the filename.

            Returns:
                ans: list
                    List of file paths
        """
        return self.selection_gen.get_file_paths(fullpath=fullpath)

    def get_file_durations(self):
        """ Get the durations of the audio files associated with this instance.

            Returns:
                ans: list
                    List of file durations in seconds
        """
        return self.selection_gen.get_file_durations()


class AudioFrameEfficientLoader(AudioFrameLoader):
    """ Load audio segments by sliding a fixed-size frame across the recording.

        AudioFrameEfficientLoader implements a more efficient approach to loading 
        overlapping audio segments and converting them to spectrograms. 
        Rather than loading and converting one frame at the time, the 
        AudioFrameEfficientLoader loads a longer frame and converts it to a 
        spectrogram which is split up into the desired shorter frames.

        Use the `num_frames` argument to specify how many frames are loaded into 
        memory at a time.

        While the segments are loaded into memory in batches, they are by default 
        returned one at a time. Use the `return_as_batch` argument to change this
        behaviour.

        Args:
            duration: float
                Segment duration in seconds. Can also be specified via the 'duration' 
                item of the 'repres' dictionary.
            step: float
                Separation between consecutive segments in seconds. If None, the step size 
                equals the segment duration. 
            path: str
                Path to folder containing .wav files. If None is specified, the current directory will be used.
            filename: str or list(str)
                relative path to a single .wav file or a list of .wav files. Optional
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            annotations: pandas DataFrame
                Annotation table. Optional.
            representation: class or list of classes
                Audio data representation. This is a class that must receive the raw audio data 
                and will transform the data into the specified audio representation object.  
                It is also possible to specify multiple audio presentations as a list. These 
                presentations must have the same duration.
            representation_params: dict or list of dict
                Dictionary containing any required and optional arguments for the representation class. If more than one
                representation is given `representation_params` must be a list of the same length and in the same order.
            num_frames: int
                Load segments in batches of size `num_frames` rather than one at the time. 
                Increasing `num_frames` can help reduce computational time.
                You can also specify `num_frames='file'` to load one wav file at the time.
            return_as_batch: bool
                Whether to return the segments individually or in batches of size `num_frames`.
                The default behaviour is to return the segments individually.
    """
    def __init__(self, duration=None, step=None, path=None, filename=None, channel=0, 
                    annotations=None, representation=Waveform, representation_params=None,
                    num_frames=12, return_as_batch=False):

        assert (isinstance(num_frames, int) and num_frames >= 1) or \
            (isinstance(num_frames, str) and num_frames.lower() == 'file'), \
            'Argument `num_frames` must be a positive integer or have the string value `file`'

        super().__init__(duration=duration, step=step, path=path, filename=filename, 
                    channel=channel, annotations=annotations, representation=representation, 
                    representation_params=representation_params)

        self.return_as_batch = return_as_batch
        self.transforms_list = []
        
        if isinstance(num_frames, int):
            self.max_batch_size = num_frames
        else:
            self.max_batch_size = np.inf

        audio_sel = next(self.selection_gen)
        self.offset = audio_sel['offset']
        self.data_dir = audio_sel['data_dir']
        self.filename = audio_sel['filename']

    def __next__(self):
        """ Load the next audio segment or batch of audio segments.

            Depending on how the loader was initialized, the return value can either be 
            an instance of :class:`BaseAudio <ketos.audio.base_audio.BaseAudio>` (or, 
            more commonly, a instance of one of its derived classes such as the 
            :class:`Waveform <ketos.audio.waveform.Waveform>` or 
            :class:`MagSpectrogram <ketos.audio.spectrogram.MagSpectrogram>`
            classes), a list of such objects, or a nested listed of such objects. 

             * If the loader was initialized with the audio representation `representation=Waveform`,   
               `representation_params=None` (default) and with `return_as_batch=False` (default), 
               the return value will be a single instance of :class:`Waveform <ketos.audio.waveform.Waveform>`.

             * If the loader was initialized with the audio representation 
               `representation=[Waveform, MagSpectrogram]`, `representation_params=[None, {'window':0.1,'step':0.02}]` 
               and with `return_as_batch=False` (default), the return value will be a list 
               of length 2, where the first entry holds an instance of 
               :class:`Waveform <ketos.audio.waveform.Waveform>` and the second entry holds an instance 
               of :class:`MagSpectrogram <ketos.audio.spectrogram.MagSpectrogram>`.

             * If the loader was initialized with the audio representation 
               `representation=[Waveform, MagSpectrogram]`, `representation_params=[None, {'window':0.1,'step':0.02}]` 
               and with `return_as_batch=True`, the return value will be a nested list with outer 
               length equal to `num_frames` and inner length 2, corresponding to the number of 
               audio representations.

            Returns: 
                : BaseAudio, list(BaseAudio), or list(list(BaseAudio))
                    Next segment or next batch of segments
        """
        if self.return_as_batch:
            self.load_next_batch()
            return self.batch
        else:           
            return self.next_in_batch()

    def next_in_batch(self):
        """ Load the next audio segment.
        
            Returns: 
                a: BaseAudio or list(BaseAudio)
                    Next audio segment
        """
        if self.counter == 0 or self.counter >= len(self.batch): 
            self.load_next_batch()
        
        a = self.batch[self.counter]
        self.counter += 1
        return a

    def load_next_batch(self):
        """ Load the next batch of audio objects.
        """
        self.batch_size = 0
        self.counter = 0
        offset = np.inf
        data_dir = self.data_dir
        filename = self.filename
        while data_dir == self.data_dir and filename == self.filename and offset > self.offset and self.batch_size < self.max_batch_size:
            self.batch_size += 1
            audio_sel = next(self.selection_gen)
            offset = audio_sel['offset']
            data_dir = audio_sel['data_dir']
            filename = audio_sel['filename']            

        duration = self.selection_gen.duration + self.selection_gen.step * (self.batch_size - 1)

        # load the data without applying transforms
        self.batch = self.load(data_dir=self.data_dir, filename=self.filename, offset=self.offset, 
            duration=duration, label=None)

        if not isinstance(self.batch, list): self.batch = [self.batch]

        # loop over the representations
        for i in range(len(self.representation)):
            transforms = self.representation_params[i]['transforms'] if 'transforms' in self.representation_params[i].keys() else []
            self.transforms_list.append(transforms)
            # segment the data
            self.batch[i] = self.batch[i].segment(window=self.selection_gen.duration, step=self.selection_gen.step)

            # apply the transforms to each segment separately 
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")        
                for j in range(len(self.batch[i])):
                    self.batch[i][j].apply_transforms(self.transforms_list[i])

        if len(self.batch) == 1: self.batch = self.batch[0]

        self.offset = offset
        self.data_dir = data_dir
        self.filename = filename


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'audio.base_audio' module within the ketos library

    This module contains the base class for the Waveform and Spectrogram classes.

    Contents:
        BaseAudio class;
        BaseAudioTimeAxis class
"""
import os
import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import ketos.audio.utils.misc as aum
from ketos.audio.annotation import AnnotationHandler, stack_annotations
from ketos.audio.utils.axis import LinearAxis


def segment_data(x, window, step=None):
    """ Divide the time axis into segments of uniform length, which may or may 
        not be overlapping.

        Window length and step size are converted to the nearest integer number 
        of time steps.

        If necessary, the data array will be padded with zeros at the end to 
        ensure that all segments have an equal number of samples. 

        Args:
            x: BaseAudioTime
                Data to be segmented
            window: float
                Length of each segment in seconds.
            step: float
                Step size in seconds.

        Returns:
            audio_objects: list(BaseAudioTime)
                Data segments
    """              
    if step is None: step = window

    time_res = x.time_res()
    win_len = aum.num_samples(window, 1. / time_res)
    step_len = aum.num_samples(step, 1. / time_res)

    # segment data array
    segs = aum.segment(x=x.data, win_len=win_len, step_len=step_len, pad_mode='zero')

    window = win_len * time_res
    step = step_len * time_res
    num_segs = segs.shape[0]

    # segment annotations
    if x.annot is not None:
        annots = x.annot.segment(num_segs=num_segs, window=window, step=step)    
    else: 
        annots = None

    # compute offsets
    offsets = np.arange(num_segs) * step

    # add global offset
    offsets += x.offset

    # create audio objects
    audio_objects = []    
    for i in range(segs.shape[0]):
        if annots is not None: annot = annots.get(id=i)
        else: annot = None
        kwargs = x.get_kwargs()
        kwargs.pop('offset', None)
        audio_objects.append(x.__class__(data=segs[i], annot=annot, offset=offsets[i], **kwargs))

    return audio_objects


class BaseAudio():
    """ Parent class for all audio classes.

        While the underlying data array can be accessed via the :attr:`data` attribute,
        it is recommended to always use the :func:`get_data` function to access the data 
        array, i.e., 

        >>> from ketos.audio.base_audio import BaseAudio
        >>> x = np.ones(6)
        >>> audio_sample = BaseAudio(data=x)
        >>> audio_sample.get_data()
        array([1., 1., 1., 1., 1., 1.])
        
        Args:
            data: numpy array
                Data
            filename: str
                Filename of the original data file, if available (optional)
            offset: float
                Position within the original data file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            duration: float
                Duration in seconds.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
            transforms: list(dict)
                List of dictionaries, where each dictionary specifies the name of 
                a transformation and its arguments, if any. For example,
                {"name":"normalize", "mean":0.5, "std":1.0}

        Attributes:
            data: numpy array
                Data 
            ndim: int
                Dimensionality of data.
            filename: str
                Filename of the original data file, if available (optional)
            offset: float
                Position within the original data file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            label: int
                Data label.
            annot: AnnotationHandler or pandas DataFrame
                AnnotationHandler object.
            allowed_transforms: dict
                Transforms that can be applied via the apply_transform method
            transform_log: list
                List of transforms that have been applied to this object
    """
    def __init__(self, data, filename='', offset=0, duration=None, label=None, annot=None, 
                    transforms=None, transform_log=None, **kwargs):

        if transform_log is None: transform_log = []
        if isinstance(annot, pd.DataFrame): annot = AnnotationHandler(annot)

        self.ndim = np.ndim(data)
        self.data = data

        self.filename = filename
        self.offset = offset
        self._duration = duration
        self.label = label

        self.annot = annot

        self.allowed_transforms = {'normalize': self.normalize, 
                                   'adjust_range': self.adjust_range}

        self.transform_log = transform_log        
        self.apply_transforms(transforms)

        self.kwargs = kwargs

    @staticmethod
    def infer_shape(**kwargs):
        """ Infers the data shape that would result if the class were 
            instantiated with a specific set of parameter values.

            Returns a None value if `duration` or `rate` are not specified.

            Args:
                duration: float
                    Duration in seconds
                rate: float
                    Sampling rate in Hz

            Returns:
                : tuple
                    Inferred shape. If the parameter value do not allow 
                    the shape be inferred, a None value is returned.
        """
        if 'duration' in kwargs.keys() and 'rate' in kwargs.keys():
            num_samples = int(kwargs['duration'] * kwargs['rate'])
            return (num_samples,)
        else:
            return None

    def get(self):
        """ Get a copy of this instance """ 
        return self.__class__(data=self.get_data(), annot=self.get_annotations(), **self.get_kwargs())

    def get_kwargs(self):
        """ Get keyword arguments required to create a copy of this instance. 

            Does not include the data array and annotation handler.    
        """
        kwargs = {}
        kwargs.update(self.get_repres_attrs())
        kwargs.update(self.get_instance_attrs())
        return kwargs

    def get_repres_attrs(self):
        """ Get audio representation attributes """ 
        attrs = {'transform_log':self.transform_log}
        return attrs

    def get_instance_attrs(self):
        """ Get instance attributes """ 
        attrs = {'filename':self.filename, 'offset':self.offset, 'duration':self._duration, 'label':self.label}
        attrs.update(self.kwargs)
        return attrs

    def get_data(self):
        """ Get underlying data.

            Returns:
                : numpy array
                    Data array 
        """
        return self.data

    def get_filename(self):
        """ Get filename.

            Returns:
                : string
                    Filename
        """
        return self.filename

    def get_offset(self):
        """ Get offset.

            Returns:
                : float
                    Offset
        """
        return self.offset

    def duration(self):
        """ Data array duration in seconds

            TODO: rename to get_duration()

            Returns:
                : float
                   Duration in seconds
        """    
        return self._duration

    def get_label(self, id=None):
        """ Get label.

            Returns:
                : int
                    Label
        """
        return self.label

    def get_annotations(self):
        """ Get annotations.

            Returns:
                : pandas DataFrame
                    Annotations 
        """
        if self.annot is None: return None
        else: return self.annot.get()

    def deepcopy(self):
        """ Make a deep copy of the present instance

            See https://docs.python.org/2/library/copy.html

            Returns:
                : BaseAudio
                    Deep copy.
        """
        return copy.deepcopy(self)

    def max(self, axis=0):
        """ Maximum data value along selected axis

            Args:
                axis: int
                    Axis along which metric is computed

            Returns:
                : array-like
                   Maximum value of the data array
        """    
        return np.max(self.data, axis=axis)

    def min(self, axis=0):
        """ Minimum data value along selected axis

            Args:
                axis: int
                    Axis along which metric is computed

            Returns:
                : array-like
                   Minimum value of the data array
        """    
        return np.min(self.data, axis=axis)

    def std(self, axis=0):
        """ Standard deviation along selected axis

            Args:
                axis: int
                    Axis along which metric is computed

            Returns:
                : array-like
                   Standard deviation of the data array
        """   
        return np.std(self.data, axis=axis) 

    def average(self, axis=0):
        """ Average value along selected axis

            Args:
                axis: int
                    Axis along which metric is computed

            Returns:
                : array-like
                   Average value of the data array
        """   
        return np.average(self.data, axis=axis)

    def median(self, axis=0):
        """ Median value along selected axis

            Args:
                axis: int
                    Axis along which metric is computed

            Returns:
                : array-like
                   Median value of the data array
        """   
        return np.median(self.data, axis=axis)

    def normalize(self, mean=0, std=1):
        """ Normalize the data array to specified mean and standard deviation.

            For the data array to be normalizable, it must have non-zero standard 
            deviation. If this is not the case, the array is unchanged by calling 
            this method. 

            Args:
                mean: float
                    Mean value of the normalized array. The default is 0.
                std: float
                    Standard deviation of the normalized array. The default is 1.
        """
        std_orig = np.std(self.data)
        if std_orig > 0:
            self.data = std * (self.data - np.mean(self.data)) / std_orig + mean
            self.transform_log.append({'name':'normalize', 'mean':mean, 'std':std})

    def adjust_range(self, range=(0,1)):
        """ Applies a linear transformation to the data array that puts the values
            within the specified range. 

            Args:
                range: tuple(float,float)
                    Minimum and maximum value of the desired range. Default is (0,1)
        """
        x_min = self.min()
        x_max = self.max()
        self.data = (range[1] - range[0]) * (self.data - x_min) / (x_max - x_min) + range[0]
        self.transform_log.append({'name':'adjust_range', 'range':range})

    def view_allowed_transforms(self):
        """ View allowed transformations for this audio object.

            Returns:
                : list
                    List of allowed transformations
        """
        return list(self.allowed_transforms.keys())

    def apply_transforms(self, transforms):
        """ Apply specified transforms to the audio object.

            Args:
                transforms: list(dict)
                    List of dictionaries, where each dictionary specifies the name of 
                    a transformation and its arguments, if any. For example,
                    {"name":"normalize", "mean":0.5, "std":1.0}

            Returns:
                None

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # read audio signal from wav file
                >>> wf = Waveform.from_wav('ketos/tests/assets/grunt1.wav')
                >>> # print allowed transforms
                >>> wf.view_allowed_transforms()
                ['normalize', 'adjust_range', 'crop', 'add_gaussian_noise', 'bandpass_filter']
                >>> # apply gaussian normalization followed by cropping
                >>> transforms = [{'name':'normalize','mean':0.5,'std':1.0},{'name':'crop','start':0.2,'end':0.7}]
                >>> wf.apply_transforms(transforms)
                >>> # inspect record of applied transforms 
                >>> wf.transform_log
                [{'name': 'normalize', 'mean': 0.5, 'std': 1.0}, {'name': 'crop', 'start': 0.2, 'end': 0.7, 'length': None}]
        """
        if transforms is None: return

        t = copy.deepcopy(transforms)
        for kwargs in t:
            name = kwargs.pop('name')
            if name in self.view_allowed_transforms():
                self.allowed_transforms[name](**kwargs)

    def annotate(self, **kwargs):
        """ Add an annotation or a collection of annotations.

            Input arguments are described in :meth:`ketos.audio.annotation.AnnotationHandler.add`
        """
        if self.annot is None: self.annot = AnnotationHandler() #if the object does not have an annotation handler, create one!

        self.annot.add(**kwargs)


class BaseAudioTime(BaseAudio):
    """ Parent class for time-series audio classes such as :class:`audio.waveform.Waveform` 
        and :class:`audio.spectrogram.Spectrogram`.

        Args:
            data: numpy array
                Data
            time_res: float
                Time resolution in seconds
            filename: str
                Filename of the original data file, if available (optional)
            offset: float
                Position within the original data file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
            transforms: list(dict)
                List of dictionaries, where each dictionary specifies the name of 
                a transformation and its arguments, if any. For example,
                {"name":"normalize", "mean":0.5, "std":1.0}

        Attributes:
            data: numpy array
                Data 
            ndim: int
                Dimensionality of data.
            time_ax: LinearAxis
                Axis object for the time dimension
            filename: str
                Filename of the original data file, if available (optional)
            offset: float
                Position within the original data file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            label: int
                Data label.
            annot: AnnotationHandler or pandas DataFrame
                AnnotationHandler object.
            allowed_transforms: dict
                Transforms that can be applied via the apply_transform method
            transform_log: list
                List of transforms that have been applied to this object
    """
    def __init__(self, data, time_res, filename='', offset=0, label=None, annot=None, 
                    transforms=None, transform_log=None, **kwargs):

        bins = max(1, data.shape[0])
        length = data.shape[0] * time_res
        self.time_ax = LinearAxis(bins=bins, extent=(0., length), label='Time (s)') #initialize time axis

        super().__init__(data=data, filename=filename, offset=offset, duration=self.duration(),
            label=label, annot=annot, transforms=transforms, transform_log=transform_log, **kwargs)

        self.allowed_transforms.update({'crop': self.crop})

    def get_repres_attrs(self):
        """ Get audio representation attributes """ 
        attrs = super().get_repres_attrs()
        attrs.update({'time_res':self.time_res()})
        return attrs

    def get_instance_attrs(self):
        """ Get instance attributes """ 
        attrs = super().get_instance_attrs()
        attrs.pop('duration', None)
        return attrs

    def time_res(self):
        """ Get the time resolution.

            Returns:
                : float
                    Time resolution in seconds
        """
        return self.time_ax.bin_width()

    def duration(self):
        """ Data array duration in seconds

            Returns:
                : float
                   Duration in seconds
        """    
        return self.time_ax.max()

    def label_array(self, label):
        """ Get an array indicating presence/absence (1/0) 
            of the specified annotation label for each time bin.

            Args:
                label: int
                    Label of interest.

            Returns:
                y: numpy.array
                    Label array
        """
        assert self.annot is not None, "An AnnotationHandler object is required for computing the label vector" 

        y = np.zeros(self.time_ax.bins)
        ans = self.annot.get(label=label)
        for _,an in ans.iterrows():
            b1 = self.time_ax.bin(an.start, truncate=True)
            b2 = self.time_ax.bin(an.end, truncate=True, closed_right=True)
            y[b1:b2+1] = 1

        return y

    def segment(self, window, step=None):
        """ Divide the time axis into segments of uniform length, which may or may 
            not be overlapping.

            Window length and step size are converted to the nearest integer number 
            of time steps.

            If necessary, the data array will be padded with zeros at the end to 
            ensure that all segments have an equal number of samples. 

            Args:
                window: float
                    Length of each segment in seconds.
                step: float
                    Step size in seconds.

            Returns:
                : list(BaseAudioTime)
                    Stacked data segments
        """   
        return segment_data(self, window, step)

    def crop(self, start=None, end=None, length=None, make_copy=False):
        """ Crop audio signal.
            
            Args:
                start: float
                    Start time in seconds, measured from the left edge of spectrogram.
                end: float
                    End time in seconds, measured from the left edge of spectrogram.
                length: int
                    Horizontal size of the cropped image (number of pixels). If provided, 
                    the `end` argument is ignored. 
                make_copy: bool
                    Return a cropped copy of the spectrogra. Leaves the present instance 
                    unaffected. Default is False.

            Returns:
                a: BaseAudio
                    Cropped data array
        """
        if make_copy:
            d = self.deepcopy()
        else:
            d = self

        # crop axis
        b1, b2 = d.time_ax.cut(x_min=start, x_max=end, bins=length)

        # crop audio signal
        d.data = d.data[b1:b2+1]

        # crop annotations, if any
        if d.annot:
            d.annot.crop(start=start, end=end)

        d.offset += d.time_ax.low_edge(0) #update time offset
        d.time_ax.zero_offset() #shift time axis to start at t=0 

        if make_copy is False:
            self.transform_log.append({'name':'crop', 'start':start, 'end':end, 'length':length})

        return d

    def plot(self, figsize=(5,4), label_in_title=True, append_title=''):
        """ Plot the data with proper axes ranges and labels.

            Optionally, also display annotations as boxes superimposed on the data.

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            Args:
                figsize: tuple
                    Figure size
                label_in_title: bool
                    Include label (if available) in figure title
                append_title: str
                    Append this string to the title
            
            Returns:
                fig: matplotlib.figure.Figure
                    A figure object.
                ax: matplotlib.axes.Axes
                    Axes object
        """
        # create canvas and axes
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize, sharex=True)

        # select the data array and attributes
        x = self.get_data()
        filename = self.get_filename()
        offset = self.get_offset()
        label = self.get_label()

        # axis labels
        ax.set_xlabel(self.time_ax.label)

        # title
        title = ""
        if filename is not None: title += "{0}".format(filename)       
        if label is not None and label_in_title:
            if len(title) > 0: title += ", "
            title += "{0}".format(label)

        title += append_title
        plt.title(title)

        # if offset is non-zero, add a second time axis at the top 
        # showing the `absolute` time
        if offset != 0:
            axt = ax.twiny()
            axt.set_xlim(offset, offset + self.duration())

        #fig.tight_layout()
        return fig, ax




import logging
import structlog
import yaml
import os
from pathlib import Path


class YAMLRenderer:
    """
    Custom renderer for structlog that outputs log entries in YAML format.
    """
    def __init__(self):
        # Use a class-level attribute to determine whether a log entry has already been written
        self.first_log_entry = True

    def __call__(self, _, __, event_dict):
        if self.first_log_entry:
            # If this is the first log entry, don't add a separator at the start
            self.first_log_entry = False
            return yaml.dump(event_dict)
        else:
            # If this is not the first log entry, add a separator at the start
            return f'---\n{yaml.dump(event_dict)}'


class KetosLogger:
    """
    Custom logger for Ketos applications.

    Attributes:
        _handler: logging.FileHandler
            A shared FileHandler used by all KetosLogger instances.

    Args:
        logger_name: str
            The name of the logger.
        log_path: str
            The full path to the desired log file. Defaults to 'ketos.log'.
        mode: str
            The mode in which the log file is opened. Defaults to 'w'.
        format: str
            Determines the format of the logs. Possible values are 'yaml' or 'json'
    """

    def __init__(self, logger_name, log_path='ketos.log', mode='w', format='yaml'):
        self.format = format

        # Ensure the output folder exists
        Path(os.path.dirname(log_path)).mkdir(parents=True, exist_ok=True)

        # Create a logger with Python's logging
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(logging.FileHandler(log_path, mode=mode))
        
        if format.lower() in ['yaml', 'yml']:
            renderer = YAMLRenderer()
        else:
            renderer = structlog.processors.JSONRenderer()

        # Configure structlog to use the standard Python logger
        structlog.configure(
            processors=[
                structlog.stdlib.add_logger_name, 
                structlog.stdlib.add_log_level, 
                structlog.processors.TimeStamper(fmt='%Y-%m-%d %H:%M:%S'),
                renderer
            ],
            context_class=dict,
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )
        
        # Now wrap the Python logger with structlog
        self.logger = structlog.wrap_logger(self.logger)

    def info(self, message, stdout=False, **kwargs):
        """
        Log an informational message.

        Args:
            message: str
                The message to log.
            stdout: bool
                Determines whether to write the message to stdout.
            **kwargs: 
                Variable length argument list to pass to the logger.
        """
        if stdout:
            print(message)
        self.logger.info(message, **kwargs)

    def error(self, message, stdout=False, **kwargs):
        """
        Log an error message.

        Args:
            message: str
                The message to log.
            stdout: bool
                Determines whether to write the message to stdout.
            **kwargs: 
                Variable length argument list to pass to the logger.
        """
        if stdout:
            print(message)
        self.logger.error(message, **kwargs)

    def exception(self, message, stdout=False, **kwargs):
        """
        Log an error message, with stack trace information.

        Args:
            message: str
                The message to log.
            stdout: bool
                Determines whether to write the message to stdout.
            **kwargs: 
                Variable length argument list to pass to the logger.
        """
        if stdout:
            print(message)
        self.logger.exception(message, **kwargs)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio.utils.misc' module within the ketos library
"""
import pytest
import unittest
import os
import numpy as np
import scipy.signal as sg
import ketos.audio.utils.misc as aum
from ketos.utils import morlet_func

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_pad_reflect_1d():
    x = np.random.rand(9)
    # default does nothing
    res = aum.pad_reflect(x, invert=True)
    assert np.all(res == x)
    # padding on left only
    res = aum.pad_reflect(x, pad_left=5, invert=True)
    assert np.all(res[5:] == x)
    assert np.all(np.flip(res[:5], axis=0) == 2*x[0] - x[1:6])
    # padding on both sides
    res = aum.pad_reflect(x, pad_left=5, pad_right=2, invert=True)
    assert np.all(res[5:-2] == x)
    assert np.all(np.flip(res[:5], axis=0) == 2*x[0] - x[1:6])
    assert np.all(np.flip(res[-2:], axis=0) == 2*x[-1] - x[-3:-1])
    # pad more than array size
    res = aum.pad_reflect(x, pad_left=12, invert=True)
    assert np.all(res[12:] == x)
    assert np.all(np.flip(res[4:12], axis=0) == 2*x[0] - x[1:9])
    assert np.all(res[:4] == 0)

def test_pad_reflect_2d():
    x = np.random.rand(9,14)
    # default does nothing
    res = aum.pad_reflect(x)
    assert np.all(res == x)
    # padding on left only
    res = aum.pad_reflect(x, pad_left=5, invert=True)
    assert np.all(res[5:] == x)
    assert np.all(np.flip(res[:5], axis=0) == 2*x[0] - x[1:6])
    # padding on both sides
    res = aum.pad_reflect(x, pad_left=5, pad_right=2, invert=True)
    assert np.all(res[5:-2] == x)
    assert np.all(np.flip(res[:5], axis=0) == 2*x[0] - x[1:6])
    assert np.all(np.flip(res[-2:], axis=0) == 2*x[-1] - x[-3:-1])

def test_num_samples():
    assert aum.num_samples(time=1.0, rate=10.) == 10
    assert aum.num_samples(time=1.1, rate=10.) == 11
    assert aum.num_samples(time=1.11, rate=10., even=True) == 12

def test_segment_args():
    # simple cases produces expected output
    args = aum.segment_args(rate=10., duration=8., offset=0., window=4., step=2.)
    assert args == {'win_len':40, 'step_len':20, 'num_segs':4, 'offset_len':-10}
    args = aum.segment_args(rate=1000., duration=1., offset=0., window=0.2, step=0.04)
    assert args == {'win_len':200, 'step_len':40, 'num_segs':25, 'offset_len':-80}
    # change in offset produces expected change in offset_len
    args = aum.segment_args(rate=10., duration=8., offset=3., window=4., step=2.)
    assert args == {'win_len':40, 'step_len':20, 'num_segs':4, 'offset_len':20}
    # window_len is always even
    args = aum.segment_args(rate=10., duration=8., offset=0., window=4.11, step=2.)
    assert args == {'win_len':42, 'step_len':20, 'num_segs':4, 'offset_len':-11}
    # step_len is always even
    args = aum.segment_args(rate=10., duration=8., offset=0., window=4., step=2.11)
    assert args == {'win_len':40, 'step_len':22, 'num_segs':4, 'offset_len':-9}
    # if the duration is not an integer multiple of the step size, we still get the expected output
    args = aum.segment_args(rate=10., duration=9., offset=0., window=4., step=2.)
    assert args == {'win_len':40, 'step_len':20, 'num_segs':5, 'offset_len':-10}

def test_segment_1d():
    x = np.arange(10)
    # check that segment with zero offset yields expected result
    res = aum.segment(x, win_len=4, step_len=2, num_segs=3, offset_len=0, )    
    ans = np.array([[0, 1, 2, 3],\
                    [2, 3, 4, 5],\
                    [4, 5, 6, 7]])
    assert np.all(ans == res)
    # check that segment with non-zero offset yields expected result
    res = aum.segment(x, win_len=4, step_len=2, num_segs=3, offset_len=-3)    
    ans = np.array([[ 3,  2,  1,  0],\
                    [ 1,  0,  1,  2],\
                    [ 1,  2,  3,  4]])
    assert np.all(ans == res)
    # check that segment with num_segs not specified yields expected result
    res = aum.segment(x, win_len=4, step_len=2)    
    ans = np.array([[0, 1, 2, 3],\
                    [2, 3, 4, 5],\
                    [4, 5, 6, 7],\
                    [6, 7, 8, 9]])
    assert np.all(ans == res)
    x = np.arange(11)
    res = aum.segment(x, win_len=4, step_len=2)    
    ans = np.array([[0, 1, 2, 3],\
                    [2, 3, 4, 5],\
                    [4, 5, 6, 7],\
                    [6, 7, 8, 9],\
                    [8, 9, 10, 9]])
    assert np.all(ans == res)

def test_segment_2d():
    x = [[n for _ in range(4)] for n in range(8)]
    x = np.array(x)
    # check that segment yields expected result
    res = aum.segment(x, num_segs=3, offset_len=0, win_len=3, step_len=2)    
    ans = np.array([[[0, 0, 0, 0],\
                     [1, 1, 1, 1],\
                     [2, 2, 2, 2]],\
                    [[2, 2, 2, 2],\
                     [3, 3, 3, 3],\
                     [4, 4, 4, 4]],\
                    [[4, 4, 4, 4],\
                     [5, 5, 5, 5],\
                     [6, 6, 6, 6]]])
    assert np.all(ans == res)

def test_segment_pass_args_as_dict():
    x = np.arange(10)
    d = {'win_len':4, 'step_len':2}
    res = aum.segment(x, **d)    
    ans = np.array([[0, 1, 2, 3],\
                    [2, 3, 4, 5],\
                    [4, 5, 6, 7],\
                    [6, 7, 8, 9]])
    assert np.all(ans == res)

def test_stft():
    # 1-s sinus signal with frequency of 100 Hz and 1 kHz sampling rate
    x = np.arange(1000)
    sig = np.sin(2 * np.pi * 100 * x / 1000) 
    # compute STFT with window length of 200, step length of 40, and offset of -(200-40)/2=-80
    seg_args = aum.segment_args(rate=1000., duration=1., offset=0., window=0.2, step=0.04)
    mag, freq_max, num_fft, seg_args, phase = aum.stft(x=sig, rate=1000, seg_args=seg_args)
    assert freq_max == 500.
    assert num_fft == 200
    assert seg_args == {'win_len':200, 'step_len':40, 'num_segs':25, 'offset_len':-80}
    assert np.all(np.argmax(mag[1:-1], axis=1) == int(100. / 500. * mag.shape[1])) #peak at f=100Hz (except first and last time bin)
    # compute STFT with window size of 0.2 s and step size of 0.04 s, and check that results are identical to above
    x = np.arange(1000)
    sig = np.sin(2 * np.pi * 100 * x / 1000) 
    mag_s, freq_max_s, num_fft_s, seg_args_s, phase = aum.stft(x=sig, rate=1000, window=0.2, step=0.04)
    assert freq_max_s == freq_max
    assert num_fft_s == num_fft
    assert seg_args_s == {'win_len':200, 'step_len':40, 'num_segs':mag_s.shape[0], 'offset_len':-80}
    assert np.all(mag_s == mag)

def test_to_decibel_returns_decibels():
    x = 7
    y = aum.to_decibel(x)
    assert y == 20 * np.log10(x) 

def test_to_decibel_can_handle_arrays():
    x = np.array([7,8])
    y = aum.to_decibel(x)
    assert np.all(y == 20 * np.log10(x))

def test_to_decibel_returns_inf_if_input_is_negative():
    x = -7
    y = aum.to_decibel(x)
    assert np.ma.getmask(y) == True

def test_spec2wave_sine_without_phase():
    ''' Check that we can reconstruct a sine waveform from a magnitude 
        spectrogram without considering the complex phase'''
    # 1-s sinus signal with frequency of 10 Hz and 1 kHz sampling rate
    x = np.arange(1000)
    sig = np.sin(2 * np.pi * 10 * x / 1000) 
    # compute STFT
    win_fun = 'hamming'
    seg_args = dict(win_len=100, step_len=20, offset_len=0)
    mag, freq_max, num_fft, seg_args, phase = aum.stft(x=sig, rate=1000, seg_args=seg_args, window_func=win_fun)
    img = aum.from_decibel(mag)
    # invert
    aud = aum.spec2wave(image=img, phase_angle=0, num_fft=num_fft, step_len=seg_args['step_len'], num_iters=50, window_func=win_fun)
    # check that recovered signal looks like a harmonic with frequency of 10 Hz
    assert sig.shape == aud.shape
    fft = np.fft.rfft(aud)
    assert np.argmax(fft) == int(10. / 500. * len(fft))

def test_spec2wave_morlet_with_phase():
    ''' Check that we can reconstruct a Morlet waveform from a magnitude 
        spectrogram considering also the complex phase'''
    # superposition of two Morlet wavelets
    sr = 20000
    x = np.linspace(start=0, stop=1, num=int(sr))
    y = morlet_func(time=x, frequency=100., width=0.1, displacement=0.5+0.3/100.) \
        + morlet_func(time=x, frequency=60., width=0.1, displacement=0.5+0.7/60.)
    # compute STFT, including complex phase
    win_fun = 'hamming'
    seg_args = dict(win_len=1000, step_len=200, offset_len=0)
    mag, freq_max, num_fft, seg_args, phase = aum.stft(x=y, rate=sr, seg_args=seg_args, window_func=win_fun, compute_phase=True)
    img = aum.from_decibel(mag)
    # recover waveform
    yr = aum.spec2wave(image=img, phase_angle=phase, num_fft=num_fft, step_len=seg_args['step_len'], num_iters=50, window_func=win_fun)
    # check that recovered signal matches the original within 1% of max amplitude
    assert yr.shape == y.shape
    assert np.all(np.abs(yr - y) < 1e-2 * np.max(np.abs(y)))
    # now, recover waveform without using phase 
    yr2 = aum.spec2wave(image=img, phase_angle=0, num_fft=num_fft, step_len=seg_args['step_len'], num_iters=50, window_func=win_fun)
    # check that recovered signal does not match the original
    assert np.any(np.abs(yr2 - y) > 0.1 * np.max(np.abs(y))) #deviation of more than 10%



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" selection_table module within the ketos library.

    This module provides functions for handling annotation tables and creating 
    selection tables. 

    A Ketos annotation table always has the column 'label'. 
    For call-level annotations, the table also contains the columns 'start' 
    and 'end', giving the start and end time of the call measured in seconds 
    since the beginning of the file. 
    The table may also contain the columns 'freq_min' and 'freq_max', giving the 
    minimum and maximum frequencies of the call in Hz, but this is not required.    
    The user may add any number of additional columns.
    Note that the table uses two levels of indices, the first index being the 
    filename and the second index being an integer to identify annotations 
    pertaining to the same file. 

    Here is a minimal example of an annotation table,

            +----------------------+-------+
            |                      | label |
            +-----------+----------+-------+
            | filename  | annot_id |       |
            +-----------+----------+-------+
            | file1.wav | 0        | 2     |
            +-----------+----------+-------+
            |           | 1        | 1     |
            +-----------+----------+-------+
            |           | 2        | 2     |
            +-----------+----------+-------+
            | file2.wav | 0        | 2     |
            +-----------+----------+-------+
            |           | 1        | 2     |
            +-----------+----------+-------+
            |           | 2        | 1     |
            +-----------+----------+-------+

    And here is a more extensive example with time information (call-level annotations) 
    and a few extra columns ('min_freq', 'max_freq' and 'file_time_stamp'),

            +----------------------+-------+------+-------+----------+----------+---------------------+
            |                      | start | end  | label | min_freq | max_freq | file_time_stamp     |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+
            | filename  | annot_id |                                                                  |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+
            | file1.wav | 0        | 7.0   | 8.1  | 2     | 180.6    | 294.3    | 2019-02-24 13:15:00 |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+
            |           | 1        | 8.5   | 12.5 | 1     | 174.2    | 258.7    | 2019-02-24 13:15:00 |
            |           +----------+-------+------+-------+----------+----------+---------------------+
            |           | 2        | 13.1  | 14.0 | 2     | 183.4    | 292.3    | 2019-02-24 13:15:00 |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+
            | file2.wav | 0        | 2.2   | 3.1  | 2     | 148.8    | 286.6    | 2019-02-24 13:30:00 |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+
            |           | 1        | 5.8   | 6.8  | 2     | 156.6    | 278.3    | 2019-02-24 13:30:00 |
            |           +----------+-------+------+-------+----------+----------+---------------------+
            |           | 2        | 9.0   | 13.0 | 1     | 178.2    | 304.5    | 2019-02-24 13:30:00 |
            +-----------+----------+-------+------+-------+----------+----------+---------------------+

    Ketos selection tables also use two level of indices. The first index is a unique, integer identifier, 
    while the second index is the filename. Moreover, selection tables always contain the columns 'start' 
    and 'end' giving the start and end time of the selection window measured in seconds since the beginning 
    of the file. This structure allows selections to span multiple files. The user may add any number of 
    additional columns to a selection table.

    Here is a minimal example of a selection table,

            +--------------------+-------+------+
            |                    | start | end  |
            +--------+-----------+-------+------+
            | sel_id | filename  |              |
            +--------+-----------+-------+------+
            | 0      | file1.wav | 1.5   | 4.5  |
            +--------+-----------+-------+------+
            | 1      | file1.wav | 12.0  | 15.0 |
            +--------+-----------+-------+------+
            |        | file2.wav | 0.0   | 5.0  |
            +--------+-----------+-------+------+
            | 2      | file2.wav | 2.0   | 10.0 |
            +--------+-----------+-------+------+
            | 3      | file2.wav | 7.0   | 15.0 |
            +--------+-----------+-------+------+
"""

import os
import warnings
import numpy as np
import pandas as pd
from ketos.utils import str_is_int, fractional_overlap
from ketos.data_handling.data_handling import find_wave_files, find_files, parse_datetime
from ketos.audio.waveform import get_duration


def unfold(table, sep=','):
    """ Unfolds rows containing multiple labels.

        Args:
            table: pandas DataFrame
                Annotation table.
            sep: str
                Character used to separate multiple labels.

        Returns:
            : pandas DataFrame
                Unfolded table
    """
    df = table    
    df = df.astype({'label': 'str'})
    s = df.label.str.split(",").apply(pd.Series, 1).stack()
    s.index = s.index.droplevel(-1)
    s.name = 'label'
    del df['label']
    df = df.join(s)
    return df


def rename_columns(table, mapper):
    """ Renames the table headings to conform with the ketos naming convention.

        Args:
            table: pandas DataFrame
                Annotation table.
            mapper: dict
                Dictionary mapping the headings of the input table to the 
                standard ketos headings.

        Returns:
            : pandas DataFrame
                Table with new headings
    """
    return table.rename(columns=mapper)


def empty_annot_table():
    """ Create an empty call-level annotation table

        Returns:
            df: pandas DataFrame
                Empty annotation table
    """
    df = pd.DataFrame(columns=['filename','label','start','end'])
    df = use_multi_indexing(df, 'annot_id')

    return df


def empty_selection_table():
    """ Create an empty selection table

        Returns:
            df: pandas DataFrame
                Empty selection table
    """
    df = pd.DataFrame(columns=['filename','label','start','end', 'annot_id'])
    df = use_multi_indexing(df, 'sel_id')
    return df


def standardize(annotations=None, sep=',', labels='auto', unfold_labels=False, 
                label_sep=',', trim_table=False, datetime_format=None, table=None, path=None):
    """ Standardize the annotation table format.

        The input table can be passed as a pandas DataFrame or as the filename of a csv file.
        The table may have either a single label per row, in which case unfold_labels should be set 
        to False, or multiple labels per row (e.g. as a comma-separated list of values), in which 
        case unfold_labels should be set to True and label_sep should be specified.

        The table headings are renamed to conform with the ketos standard naming convention, following the 
        name mapping specified by the user. 

        Note that the standardized output table has two levels of indices, the first index being the 
        filename and the second index the annotation identifier. 

        The label mapping is stored as a class attribute named 'label_dict' within the output table 
        and may be retrieved with `df.attrs['label_dict']`.

        Required Columns:
            - 'filename': The name or path of the file associated with each annotation.
            - 'label': The label or category associated with each annotation. If `unfold_labels` is True,
                    this column may contain multiple labels separated by `label_sep`.

            Optional Columns (depending on usage):
            - 'start': The start time or position of the annotation.
            - 'end': The end time or position of the annotation.

        Args:
            annotations: str, pandas DataFrame
                If a string, it is assumed to be the path to a CSV file containing the annotation table.
                If a pandas DataFrame, it is used directly as the annotation table.
            sep: str
                Separator. Only relevant if filename is specified. Default is ",".
            labels: 'auto', None, dict, or list
                - 'auto' (default): All unique labels in the table are automatically mapped to integers starting from 0.
                - None: No label mapping is applied, labels are left as-is.
                - dict: A user-specified mapping of labels to integers (Note that ketos expects labels to be incremental integers starting with 0).
                - list: A subset of labels to map to integers starting from 0.
                Any unspecified label is mapped to -1.
            unfold_labels: bool
                Should be set to True if any of the rows have multiple labels and False otherwise (default).
            label_sep: str
                Character used to separate multiple labels. Only relevant if unfold_labels is set to True. Default is ",".
            trim_table: bool
                Keep only the columns prescribed by the Ketos annotation format and any additional columns specified 
                in the mapper dictionary.
            datetime_format: str
                String defining the date-time format. 
                Example: %d_%m_%Y* would capture "14_3_1999.txt".
                See https://pypi.org/project/datetime-glob/ for a list of valid directives.
                If specified, the method will look for a column named 'datetime' and, if found, attempt to parse the 
                values in this column. If your datetime column has a different name, use the `mapper` argument to change 
                its name to 'datetime'. If the method does not find a column named 'datetime' it will attempt to 
                parse the datetime information from the filename column.
            table: pandas DataFrame (deprecated)
                Deprecated. Use 'annotations' instead.
            path: str (deprecated)
                Deprecated. Use 'annotations' instead.

        Returns:
            df: pandas DataFrame
                Standardized annotation table
        
        Note:
            The function assumes that the necessary preprocessing (e.g., renaming columns to match the expected names) 
            has been done prior to calling this function. It is the responsibility of the user to ensure that the input 
            table conforms to the expected format.
    """
    if table is not None or path is not None:
        warnings.warn("The 'table' and 'path' arguments are deprecated and will be removed in a future version. "
                      "Use the 'annotations' argument instead.", DeprecationWarning)
        
        if table is not None:
            annotations = table
        elif path is not None:
            annotations = path
    # Determine if 'data' is a file path or a DataFrame
    if isinstance(annotations, str):
        df = pd.read_csv(annotations, sep=sep)
    elif isinstance(annotations, pd.DataFrame):
        df = annotations
    else:
        raise ValueError("Annotations must be a pandas DataFrame or a file path to a CSV file.")

    label_mapping = dict()
    # Handle label mapping based on the labels argument
    if labels == 'auto':
        unique_labels = sorted(pd.unique(df['label']))
        label_mapping = {label: i for i, label in enumerate(unique_labels)}
        df['label'] = df['label'].map(label_mapping).fillna(-1)
    elif isinstance(labels, list):
        label_mapping = {label: i for i, label in enumerate(labels)}
        df['label'] = df['label'].map(label_mapping).fillna(-1)
    elif isinstance(labels, dict):
        label_mapping = labels
        df['label'] = df['label'].map(label_mapping).fillna(-1)
    elif labels is not None:
        raise ValueError("Unsupported value for labels argument. Use 'auto', None, a list of labels, or a dict.")
    
    # keep only relevant columns
    if trim_table:
        df = trim(df)
    
    # Ensure the DataFrame contains 'filename' and 'label' columns
    required_columns = ['filename', 'label']
    missing_columns = [col for col in required_columns if col not in df.columns]
    assert not missing_columns, f"Missing required column(s): {', '.join(missing_columns)}"

    if unfold_labels:
        df = unfold(df, sep=label_sep)

    # always sort by filename (first) and start time (second)
    by = ['filename']
    if 'start' in df.columns.values: 
        by += ['start']
    df.sort_values(by=by, inplace=True, ignore_index=True)

    # parse datetime field
    if datetime_format is not None:
        if 'datetime' in df.columns.values:
            df['datetime'] = df['datetime'].apply(lambda x: parse_datetime(x, fmt=datetime_format))
        else:
            df['datetime'] = df.apply(lambda x: parse_datetime(os.path.basename(x.filename), fmt=datetime_format), axis=1)
    
    # transform to multi-indexing
    df = use_multi_indexing(df, 'annot_id')

    # store label dictionary as class attribute
    df.attrs["label_dict"] = label_mapping
    
    # enforce float for select columns
    float_cols = ['start','end','freq_min','freq_max']
    for c in float_cols:
        if c in df.columns.values:
            df[c] = df[c].astype(float)
        
    return df


def use_multi_indexing(df, level_1_name):
    """ Change from single-level indexing to double-level indexing. 
        
        The first index level is the filename while the second 
        index level is a cumulative integer.

        Args:
            table: pandas DataFrame
                Singly-indexed table. Must contain a column named 'filename'. 

        Returns:
            table: pandas DataFrame
                Multi-indexed table.
    """
    df = df.set_index([df.filename, df.index])
    df = df.drop(['filename'], axis=1)
    df = df.sort_index()
    df.index = pd.MultiIndex.from_arrays(
        [df.index.get_level_values(0), df.groupby(level=0).cumcount()],
        names=['filename', level_1_name])

    return df


def trim(table, extra_cols=None):
    """ Keep only the columns prescribed by the Ketos annotation format.

        Args:
            table: pandas DataFrame
                Annotation table. 
            extra_cols: list(str)
                Any additional columns that we wish to keep

        Returns:
            table: pandas DataFrame
                Annotation table, after removal of columns.
    """
    keep_cols = ['filename', 'label', 'start', 'end', 'freq_min', 'freq_max']
    if extra_cols is not None:
        keep_cols += extra_cols
    drop_cols = [x for x in table.columns.values if x not in keep_cols]
    table = table.drop(drop_cols, axis=1)
    return table

def is_standardized(table, has_time=False, verbose=True):
    """ Check if the table has the correct indices and the minimum required columns.

        Args:
            table: pandas DataFrame
                Annotation table. 
            has_time: bool
                Require time information for each annotation, i.e. start and stop times.
            verbose: bool
                If True and the table is not standardized, print a message with an example table in the standard format.

        Returns:
            res: bool
                True if the table has the standardized Ketos format. False otherwise.
    """
    required_indices = ['filename', 'annot_id']
    required_cols = ['label']
    if has_time:
        required_cols = required_cols + ['start', 'end']

    mis_cols = [x for x in required_cols if x not in table.columns.values]
    res = (table.index.names == required_indices) and (len(mis_cols) == 0)

    message = """ Your table is not in the Ketos format.

            It should have two levels of indices: filename and annot_id.
            It should also contain at least the 'label' column.
            If your annotations have time information, these should appear in the 'start' and 'end' columns

            extra columns are allowed.

            Here is a minimum example:

                                 label
            filename  annot_id                    
            file1.wav 0          2
                      1          1
                      2          2
            file2.wav 0          2
                      1          2
                      2          1


            And here is a table with time information and a few extra columns ('min_freq', 'max_freq' and 'file_time_stamp')

                                 start   end  label  min_freq  max_freq  file_time_stamp
            filename  annot_id                    
            file1.wav 0           7.0   8.1      2    180.6     294.3    2019-02-24 13:15:00
                      1           8.5  12.5      1    174.2     258.7    2019-02-24 13:15:00
                      2          13.1  14.0      2    183.4     292.3    2019-02-24 13:15:00
            file2.wav 0           2.2   3.1      2    148.8     286.6    2019-02-24 13:30:00
                      1           5.8   6.8      2    156.6     278.3    2019-02-24 13:30:00
                      2           9.0  13.0      1    178.2     304.5    2019-02-24 13:30:00

    
    """
    if res == False and verbose == True:
        print(message)            

    return res


def _create_label_dict(labels, discard_labels=None, start_labels_at_1=False):
    """ Create label dictionary, following the convetion:

            * signal_labels are mapped to 1,2,3,...
            * backgr_labels are mapped to 0
            * discard_labels are mapped to -1

        Args:
            signal_labels: list, or list of lists
                Labels of interest. Will be mapped to 1,2,3,...
                Several labels can be mapped to the same integer by using nested lists. For example, 
                signal_labels=[A,[B,C]] would result in A being mapped to 1 and B and C both being mapped 
                to 2.
            backgr_labels: list
                Labels will be grouped into a common "background" class (0).
            discard_labels: list
                Labels will be grouped into a common "discard" class (-1).
            start_labels_at_1: bool
                Map labels to 1,2,3,... instead of 0,1,2,... Default is False.

        Returns:
            label_dict: dict
                Dict that maps old labels to new labels.

        This function is deprecated and will be removed in a future version.
        Users should transition to creating their own label mapping.
    """
    # Deprecation warning to alert users
    warnings.warn(
        "_create_label_dict is deprecated and will be removed in a future version. "
        "Users should transition to creating their own label mapping.",
        DeprecationWarning
    )

    label_dict = dict()  
    if discard_labels is not None:  
        for l in discard_labels: label_dict[l] = -1
    
    num = 1 if start_labels_at_1 else 0
    for l in labels:
        if isinstance(l, list):
            for ll in l:
                label_dict[ll] = num

        else:
            label_dict[l] = num

        num += 1

    return label_dict


def label_occurrence(table):
    """ Identify the unique labels occurring in the table and determine how often 
        each label occurs.

        The input table must have the standardized Ketos format, see 
        :func:`data_handling.selection_table.standardize`. In particular, each 
        annotation should have only a single label value.

        Args:
            table: pandas DataFrame
                Input table.

        Results:
            occurrence: dict
                Dictionary where the labels are the keys and the values are the occurrences.
    """
    occurrence = table.groupby('label').size().to_dict()
    return occurrence

def cast_to_str(labels, nested=False):
    """ Convert every label to str format. 

        If nested is set to True, a flattened version of the input 
        list is also returned.

        Args:
            labels: list
                Input labels
            nested: bool
                Indicate if the input list contains (or may contain) sublists.
                False by default. If True, a flattened version of the 
                list is also returned.

        Results:
            labels_str: list
                Labels converted to str format
            labels_str_flat: list
                Flattened list of labels. Only returned if nested is set to True.
    """
    if not nested:
        labels_str = [str(x) for x in labels]
        return labels_str

    else:
        labels_str = []
        labels_str_flat = []
        for x in labels:
            if isinstance(x, list):
                sublist = []
                for xx in x:
                    labels_str_flat.append(str(xx))
                    sublist.append(str(xx))

                labels_str.append(sublist)

            else:
                labels_str_flat.append(str(x))
                labels_str.append(str(x))

        return labels_str, labels_str_flat

def select(annotations, length, step=0, min_overlap=0, center=False, discard_long=False, 
    keep_id=False, keep_freq=False, label=None, avoid_label=None, discard_outside=False, files=None):
    """ Generate a selection table by defining intervals of fixed length around 
        annotated sections of the audio data. Each selection created in this 
        way is characterized by a single, integer-valued, label.

        This approach to generating selections lends itself well to cases in which 
        the annotated sections are well separated and rarely overlap. If this is not 
        the case, you may find the related function :func:`data_handling.selection_table.select_by_segmenting` 
        more useful.

        By default all annotated sections are used for generating selections except 
        those with label -1 which are ignored. Use the `label` argument to only 
        generate selections for specific labels. 

        Conversely, the argument `avoid_label` can be used to ensure that the generated 
        selections do not overlap with annotated sections with specific labels. For example, 
        if `label=[1,2]` and `avoid_label=[4]`, selections will be generated for every 
        annotated section with label 1 or 2, but any selection that happens to overlap with 
        an annotated sections with label 4 will be discarded. 

        The input table must have the standardized Ketos format and contain call-level 
        annotations, see :func:`data_handling.selection_table.standardize`.

        The output table uses two levels of indexing, the first level being the 
        filename and the second level being a selection id.

        The generated selections have uniform length given by the `length` argument. Annotated 
        sections longer than the specified length will be cropped (unless discard_long=True) 
        whereas shorter sections will be extended to achieve the specified length. 

        The `step` and `min_overlap` arguments may be used to generate multiple, time-shifted 
        selections for every annotated sections.

        Note that the selections may have negative start times and/or end times 
        that exceed the file duration, unless discard_outside=True in which case only 
        selections with start times and end times within the file duration are returned.

        Args:
            annotations: pandas DataFrame
                Input table with call-level annotations.
            length: float
                Selection length in seconds.
            step: float
                Produce multiple selections for each annotated  section by shifting the selection 
                window in steps of length step (in seconds) both forward and backward in 
                time. The default value is 0.
            min_overlap: float
                Minimum required overlap between the selection and the annotated section, expressed 
                as a fraction of whichever of the two is shorter. Only used if step > 0. 
            center: bool
                Center annotations. Default is False.
            discard_long: bool
                Discard all annotations longer than the output length. Default is False.
            keep_id: bool
                For each generated selection, include the id of the annotation from which 
                the selection was generated.
            keep_freq: bool
                For each generated selection, include the min and max frequency, if known.
            label: int or list(int)
                Only create selections for annotated sections with these labels.
            avoid_label: int, list(int) or str
                Avoid overlap with annotated sections with these labels. If overlap is to be avoided 
                with all other labels but the labels specified by the `label` argument, set `avoid_label="ALL"`.
            discard_outside: bool
                Discard selections that extend beyond file duration. Requires that a file duration 
                table is specified via the `files` argument.
            files: pandas DataFrame
                Table with file durations in seconds. Must contain columns named 'filename' and 'duration'.
                Only required if `discard_outside=True`.

        Results:
            df: pandas DataFrame
                Output selection table.

        Example:
            >>> import pandas as pd
            >>> from ketos.data_handling.selection_table import select, standardize
            >>> 
            >>> #Load and inspect the annotations.
            >>> df = pd.read_csv("ketos/tests/assets/annot_001.csv")
            >>>
            >>> #Standardize annotation table format
            >>> df = standardize(df, labels={0:1, 1:2})
            >>> print(df)
                                start   end  label
            filename  annot_id                    
            file1.wav 0           7.0   8.1      2
                      1           8.5  12.5      1
                      2          13.1  14.0      2
            file2.wav 0           2.2   3.1      2
                      1           5.8   6.8      2
                      2           9.0  13.0      1
            >>> 
            >>> #Create a selection table by defining intervals of fixed 
            >>> #length around every annotation.
            >>> #Set the length to 3.0 sec and require a minimum overlap of 16%
            >>> #between selection and annotations.
            >>> #Also, create multiple time-shifted versions of the same selection
            >>> #using a step size of 1.0 sec.     
            >>> df_sel = select(df, length=3.0, step=1.0, min_overlap=0.16, center=True, keep_id=True) 
            >>> print(df_sel.round(2))
                              label  start    end  annot_id
            filename  sel_id                               
            file1.wav 0           2   5.05   8.05         0
                      1           1   6.00   9.00         1
                      2           2   6.05   9.05         0
                      3           1   7.00  10.00         1
                      4           2   7.05  10.05         0
                      5           1   8.00  11.00         1
                      6           1   9.00  12.00         1
                      7           1  10.00  13.00         1
                      8           1  11.00  14.00         1
                      9           2  11.05  14.05         2
                      10          1  12.00  15.00         1
                      11          2  12.05  15.05         2
                      12          2  13.05  16.05         2
            file2.wav 0           2   0.15   3.15         0
                      1           2   1.15   4.15         0
                      2           2   2.15   5.15         0
                      3           2   3.80   6.80         1
                      4           2   4.80   7.80         1
                      5           2   5.80   8.80         1
                      6           1   6.50   9.50         2
                      7           1   7.50  10.50         2
                      8           1   8.50  11.50         2
                      9           1   9.50  12.50         2
                      10          1  10.50  13.50         2
                      11          1  11.50  14.50         2
                      12          1  12.50  15.50         2
    """
    if len(annotations) == 0:
        return empty_selection_table()

    df = annotations.copy()
    df['annot_id'] = df.index.get_level_values(1)

    # check that input table has expected format
    assert is_standardized(df, has_time=True), 'Annotation table appears not to have the expected structure.'

    # select labels
    if label is not None:
        if isinstance(label, int): label = [label]
        df = df[df['label'].isin(label)]

    # discard annotations with label -1
    df = df[df['label'] != -1]

    # compute length of every annotation
    df['length'] = df['end'] - df['start']

    # discard annotations longer than the requested length
    if discard_long:
        df = df[df['length'] <= length]

    # We need to ensure that the annotation is valid, that is, the start time must be smaller than end time. 
    # Otherwise we skip (remove) the annotation and throw a warning
    negative_length = df[df['length'] < 0] # select rows with the issue to issue warnings
    
    if (len(negative_length.index) > 0):
        df = df[df['length'] >= 0] # remove the rows from the dataframe

        for idx,row in negative_length.iterrows():
            warnings.warn("File {0}, annotation {1} has a start time ({2}) greater than end time ({3}). Skipping annotation".format(idx[0], idx[1], row['start'], row['end']), category=UserWarning, stacklevel=2)  

    # number of annotations
    N = len(df)

    # alignment of new annotations relative to original ones
    if center:
        df['start_new'] = df['start'] + 0.5 * (df['length'] - length)
    else:
        df['start_new'] = df['start'] + np.random.random_sample(N) * (df['length'] - length)

    # create multiple time-shited instances of every annotation
    if step > 0:
        df_new = None
        for idx,row in df.iterrows():
            t = row['start_new']
            df_shift = time_shift(annot=row, time_ref=t, length=length, min_overlap=min_overlap, step=step)
            df_shift['filename'] = idx[0]

            if df_new is None:
                df_new = df_shift
            else:
                df_new = pd.concat([df_new, df_shift])

        # sort by filename and offset
        df = df_new.sort_values(by=['filename','start_new'], axis=0, ascending=[True,True]).reset_index(drop=True)

        # transform to multi-indexing
        df = use_multi_indexing(df, 'sel_id')

    # rename index
    df.index.rename('sel_id', level=1, inplace=True) 

    # drop old/temporary columns, and rename others
    df = df.drop(['start', 'end', 'length'], axis=1)
    df = df.rename(columns={"start_new": "start"})
    df['end'] = df['start'] + length

    # keep annotation id
    if not keep_id:
        df = df.drop(columns=['annot_id'])
    else:
        # re-order columns so annot_it appears last
        cols = df.columns.values.tolist()
        p = cols.index('annot_id')
        cols_new = cols[:p] + cols[p+1:] + ['annot_id']
        df = df[cols_new]
        df = df.astype({'annot_id': int}) #ensure annot_id is int

    # ensure label is integer
    df = df.astype({'label':int})

    # discard selections that overlap with unwanted annotations
    if avoid_label is not None:

        if isinstance(avoid_label, int): 
            avoid_label = [avoid_label]

        elif isinstance(avoid_label, str) and avoid_label.lower() == "all" and label is not None:
            labels = pd.unique(annotations.label)
            avoid_label = labels[~np.isin(labels,label)]
            avoid_label = avoid_label.tolist()

        if isinstance(avoid_label, list):
            def func(y, start, end, label):
                y = y[(y.label.isin(label)) & (y.end>=start) & (y.start<=end)]
                return len(y)
            
            df['overlap'] = df.apply(lambda x: func(annotations.loc[x.name[0]], label=avoid_label, start=x.start, end=x.end), axis=1)
            df = df[df['overlap']==0]
            df = df.drop(columns=['overlap'])
        
    # discard selections that extend beyond duration of file
    if discard_outside:
        if files is None:
            warnings.warn("discard_outside=True requires files to be specified")
        else:
            files = files.set_index('filename')
            df['outside'] = df.apply(lambda x: (x.start < 0) or (x.end > files.loc[x.name[0]].duration), axis=1)
            df = df[df['outside']==False]
            df = df.drop(columns=['outside'])

    if not keep_freq:
        df = df.drop(columns=['freq_min','freq_max'], errors='ignore')

    return df

def time_shift(annot, time_ref, length, step, min_overlap):
    """ Create multiple instances of the same selection by stepping in time, both 
        forward and backward.

        The time-shifted instances are returned in a pandas DataFrame with the same columns as the 
        input annotation, plus a column named 'start_new' containing the start times 
        of the shifted instances.

        Args:
            annot: pandas Series or dict
                Reference annotation. Must contain the labels/keys 'start' and 'end'.
            time_ref: float
                Reference time used as starting point for the stepping.
            length: float
                Output annotation length in seconds.
            step: float
                Produce multiple instances of the same selection by shifting the annotation 
                window in steps of length step (in seconds) both forward and backward in 
                time. The default value is 0.
            min_overlap: float
                Minimum required overlap between the selection intervals and the original 
                annotation, expressed as a fraction of whichever is smaller, the annotation 
                duration or the selection length.   

        Results:
            df: pandas DataFrame
                Output annotation table. The start times of the time-shifted annotations are 
                stored in the column 'start_new'.

        Example:
            >>> import pandas as pd
            >>> from ketos.data_handling.selection_table import time_shift
            >>> 
            >>> #Create a single 2-s long annotation
            >>> annot = {'filename':'file1.wav', 'label':1, 'start':12.0, 'end':14.0}
            >>>
            >>> #Step across this annotation with a step size of 0.2 s, creating 1-s long annotations that 
            >>> #overlap by at least 50% with the original 2-s annotation 
            >>> df = time_shift(annot, time_ref=13.0, length=1.0, step=0.2, min_overlap=0.5)
            >>> print(df.round(2))
                filename  label  start   end  start_new
            0  file1.wav      1   12.0  14.0       11.6
            1  file1.wav      1   12.0  14.0       11.8
            2  file1.wav      1   12.0  14.0       12.0
            3  file1.wav      1   12.0  14.0       12.2
            4  file1.wav      1   12.0  14.0       12.4
            5  file1.wav      1   12.0  14.0       12.6
            6  file1.wav      1   12.0  14.0       12.8
            7  file1.wav      1   12.0  14.0       13.0
            8  file1.wav      1   12.0  14.0       13.2
            9  file1.wav      1   12.0  14.0       13.4
    """
    if length <= 0:
        raise AssertionError("Length must be positive and greater than zero, found {0}".format(length))

    if isinstance(annot, dict):
        row = pd.Series(annot)
    elif isinstance(annot, pd.Series):
        row = annot.copy()
    
    row['start_new'] = time_ref
    rows_new = [row]

    # step backwards and forwards
    for sign in [-1, 1]:
        counter = 1
        while True:
            t0 = time_ref + sign * counter * step
            o = fractional_overlap(a=(t0,t0+length), b=(row['start'], row['end']))
            if o < min_overlap: 
                break
            ri = row.copy()
            ri['start_new'] = t0
            rows_new.append(ri)
            counter += 1

    # create DataFrame
    df = pd.DataFrame(rows_new)

    # sort according to new start time
    df = df.sort_values(by=['start_new'], axis=0, ascending=[True]).reset_index(drop=True)

    return df

def file_duration_table(path, search_subdirs=False, datetime_format=None):
    """ Create file duration table.

        Args:
            path: str
                Path to folder with audio files with extensions wav, WAV, flac, FLAC.
            search_subdirs: bool
                If True, search include also any audio files in subdirectories.
                Default is False.
            datetime_format: str
                String defining the date-time format. 
                Example: %d_%m_%Y* would capture "14_3_1999.txt".
                See https://pypi.org/project/datetime-glob/ for a list of valid directives.
                If specified, the method will attempt to parse the datetime information from the filename.

        Returns:
            df: pandas DataFrame
                File duration table. Columns: filename, duration, (datetime)
    """
    paths = find_files(path=path, return_path=True, search_subdirs=search_subdirs, substr=['.wav', '.WAV', '.flac', '.FLAC'])
    durations = get_duration([os.path.join(path,p) for p in paths])
    df = pd.DataFrame({'filename':paths, 'duration':durations})
    if datetime_format is None:
        return df

    df['datetime'] = df.apply(lambda x: parse_datetime(os.path.basename(x.filename), fmt=datetime_format), axis=1)
    return df

def create_rndm_selections(files, length, num, label=0, annotations=None, no_overlap=False, trim_table=False, buffer=0):
    """ Create selections of uniform length, randomly distributed across the 
        data set and not overlapping with any annotations. The created selections
        will have a label value defined by the 'label' parameter.

        The random sampling is performed without regard to already created 
        selections. Therefore, it is in principle possible that some of the created 
        selections will overlap, although in practice this will only occur with very 
        small probability, unless the number of requested selections (num) is very 
        large and/or the (annotation-free part of) the data set is small in size.

        To avoid any overlap, set the 'no_overlap' to True, but note that this can 
        lead to longer execution times.

        Use the 'buffer' argument to ensure a minimum separation between
        selections and the annotated segments. This can be useful if the annotation 
        start and end times are not always fully accurate.

        Args:
            files: pandas DataFrame
                Table with file durations in seconds. 
                Should contain columns named 'filename' and 'duration'.
            length: float
                Selection length in seconds.
            num: int
                Number of selections to be created.
            label: int
                Value to be assigned to the created selections.
            annotations: pandas DataFrame
                Annotation table. Optional.
            no_overlap: bool
                If True, randomly selected segments will have no overlap.
            trim_table: bool
                Keep only the columns prescribed by the Ketos annotation format.
            buffer: float
                Minimum separation in seconds between the background selections and the annotated segments.
                The default value is zero.

        Returns:
            table_backgr: pandas DataFrame
                Output selection table.

        Example:
            >>> import pandas as pd
            >>> import numpy as np
            >>> from ketos.data_handling.selection_table import select
            >>> 
            >>> #Ensure reproducible results by fixing the random number generator seed.
            >>> np.random.seed(3)
            >>> 
            >>> #Load and inspect the annotations.
            >>> df = pd.read_csv("ketos/tests/assets/annot_001.csv")
            >>> print(df)
                filename  start   end  label
            0  file1.wav    7.0   8.1      1
            1  file1.wav    8.5  12.5      0
            2  file1.wav   13.1  14.0      1
            3  file2.wav    2.2   3.1      1
            4  file2.wav    5.8   6.8      1
            5  file2.wav    9.0  13.0      0
            >>>
            >>> #Standardize annotation table format
            >>> df = standardize(annotations=df, labels={0:1, 1:2})  # Standardize annotation table format (we want to create background random segments with label 0, so lets map the labels we have to 1 and 2)
            >>> print(df)
                                start   end  label
            filename  annot_id                    
            file1.wav 0           7.0   8.1      2
                      1           8.5  12.5      1
                      2          13.1  14.0      2
            file2.wav 0           2.2   3.1      2
                      1           5.8   6.8      2
                      2           9.0  13.0      1
            >>>
            >>> #Enter file durations into a pandas DataFrame
            >>> file_dur = pd.DataFrame({'filename':['file1.wav','file2.wav','file3.wav',], 'duration':[18.,20.,15.]})
            >>> 
            >>> #Create randomly sampled background selection with fixed 3.0-s length.
            >>> df_bgr = create_rndm_selections(annotations=df, files=file_dur, length=3.0, num=12, trim_table=True) 
            >>> print(df_bgr.round(2))
                              start    end  label
            filename  sel_id                     
            file1.wav 0        3.38   6.38      0
                      1        3.89   6.89      0
            file2.wav 0       16.52  19.52      0
            file3.wav 0        0.29   3.29      0
                      1        2.77   5.77      0
                      2        3.23   6.23      0
                      3        5.49   8.49      0
                      4        5.63   8.63      0
                      5        6.69   9.69      0
                      6        6.71   9.71      0
                      7        8.18  11.18      0
                      8       10.33  13.33      0
    """
    if len(files) == 0:
        return empty_selection_table()

    assert isinstance(label, int), 'label is not int. Found {0}'.format(type(label))

    # compute lengths, and discard segments shorter than requested length
    c = files[['filename','duration']]

    if 'offset' in files.columns.names: c['offset'] = files['offset']
    else: c['offset'] = 0

    c.reset_index(drop=True, inplace=True)
    c['length'] = c['duration'] - length
    c = c[c['length'] >= 0]

    start_list, end_list, filename_list = [], [], []
    # Converting data from pandas df to lists and numpy array to use inside while loop (Much more efficience than accessing pandas row by row)
    durations = files['duration'].to_numpy()
    lengths = c['length'].to_numpy()
    offsets = c['offset'].to_numpy()
    filenames = c['filename'].tolist()

    cnt = 0
    probabilities = durations/durations.sum()

    # randomply sample
    # Loop until we achieve the desired number of samples
    while cnt < num:
        # Randomly choose a file to sample from
        # We want to sample from files with a longer duration with a higher probability than files with a lower duration.
        # Create array of indexes of size the number of segments to generate
        indices = np.random.choice(len(c), size=num, replace=True, p=probabilities)

        # Explanation of the following loop: The reason of using a nested for loop here is purely for efficiency.
        # It could be removed and instead of generating an array of indexes we could reandomly sample one index at a time.
        # However, random.choice is the most computationaly expensive operation here with O(n + n log m ) when p is specified
        # By reducing the amount of times we have to call it we are drastically reducing computaitonal time.
        for idx in indices:
            # Randomly sample a segment of duration = length from the timeseries
            t = np.random.random_sample() * lengths[idx]
            start = t + offsets[idx]
            end   = start + length
            fname = filenames[idx]

            # If given, gheck if the sampled segment does not overlap with an annotation
            if annotations is not None:
                q = query(annotations, filename=fname, start=start-buffer, end=end+buffer)
                if len(q) > 0: continue
            
            # If set, check if segments do not overlap with each other
            if no_overlap and cnt > 0:
                # Query requires passing a df as the first argument, therefore lets create a temporary df
                tmp_df = pd.DataFrame({'start': start_list, 'end': end_list, 'filename': filename_list})
                q = query(tmp_df.set_index(tmp_df.filename), filename=fname, start=start, end=end)
                if len(q) > 0: continue

            start_list.append(start)
            end_list.append(end)
            filename_list.append(fname)
            cnt += 1

            if cnt == num:
                break

    # Create Pandas df at the end
    df = pd.DataFrame({'start': start_list, 'end': end_list, 'filename': filename_list})
    # We want to keep all columns that were present in the files dataframe. This is a classic inner join case on the filename
    df = pd.DataFrame.merge(df, files, on='filename')
    
    # sort by filename and offset
    df = df.sort_values(by=['filename','start'], axis=0, ascending=[True,True]).reset_index(drop=True)

    # re-order columns
    col_names = ['filename','start','end']
    if not trim_table:
        names = df.columns.values.tolist()
        for name in col_names: names.remove(name)
        col_names += names

    df = df[col_names]

    df['label'] = label #add label

    # transform to multi-indexing
    df = use_multi_indexing(df, 'sel_id')

    return df

def random_choice(df, siz):
    """ Randomly select a specified number of elements from a table.

        Args:
            df: pandas DataFrame
                Selection table or annotation table
            siz: int
                Number of elements to be selected

        Returns:
            sel: pandas DataFrame
                Reduced table
    """

    name_id = 'sel_id' if 'sel_id' in df.index.names else 'annot_id'
    n = min(siz, len(df))
    idx = np.sort(np.random.choice(np.arange(len(df), dtype=int), size=n, replace=False)).tolist()
    df = df.reset_index()
    df = df.loc[idx]
    df = df.set_index([df.filename, df[name_id]])
    df = df.drop(['filename', name_id], axis=1)
    df = df.sort_index()
    return df

def select_by_segmenting(files, length, annotations=None, step=None,
    pad=True, discard_empty=False, keep_only_empty=False, label_empty=0, 
    avoid_label=None):
    """ Generate a selection table by stepping across the audio files, using a fixed 
        step size (step) and fixed selection window size (length). 
        
        Unlike the :func:`data_handling.selection_table.select` method, selections 
        created by this method are not characterized by a single, integer-valued 
        label, but rather a list of annotations (which can have any length, including zero).

        Therefore, the method returns not one, but two tables: A selection table indexed by 
        filename and segment id, and an annotation table indexed by filename, segment id, 
        and annotation id.

        However, if `keep_only_empty=True` only a selection table is returned. This table 
        has a column named `label` with all entries having the same value, as specified via
        the `label_empty` argument.

        Args:
            files: pandas DataFrame
                Table with file durations in seconds. 
                Should contain columns named 'filename' and 'duration'.
            length: float
                Selection length in seconds.
            annotations: pandas DataFrame
                Annotation table.
            step: float
                Selection step size in seconds. If None, the step size is set 
                equal to the selection length.
            pad: bool
                If True (default), the last selection window is allowed to extend 
                beyond the endpoint of the audio file.
            discard_empty: bool
                If True, only selection that contain annotations will be used. 
                If False (default), all selections are used.
            keep_only_empty: bool
                If True, only selections *without* any annotations are used, and 
                only the selections table is returned. Default is False. 
            label_empty: int
                Only relevant if keep_only_empty is True. Value to be assigned to
                selections without annotations. Default is 0.
            avoid_label: int or list(int)
                If specified, only selections without annotations with these labels 
                are used.

        Returns:
            sel: pandas DataFrame
                Selection table
            annot: pandas DataFrame
                Annotations table. Only returned if annotations is specified 
                and keep_only_empty is False.

        Example:
            >>> import pandas as pd
            >>> from ketos.data_handling.selection_table import select_by_segmenting, standardize
            >>> 
            >>> #Load and inspect the annotations.
            >>> annot = pd.read_csv("ketos/tests/assets/annot_001.csv")
            >>>
            >>> #Standardize annotation table format
            >>> annot = standardize(annot, labels={0:1, 1:2})
            >>> print(annot)
                                start   end  label
            filename  annot_id                    
            file1.wav 0           7.0   8.1      2
                      1           8.5  12.5      1
                      2          13.1  14.0      2
            file2.wav 0           2.2   3.1      2
                      1           5.8   6.8      2
                      2           9.0  13.0      1
            >>>
            >>> #Create file table
            >>> files = pd.DataFrame({'filename':['file1.wav', 'file2.wav', 'file3.wav'], 'duration':[11.0, 19.2, 15.1]})
            >>> print(files)
                filename  duration
            0  file1.wav      11.0
            1  file2.wav      19.2
            2  file3.wav      15.1
            >>>
            >>> #Create a selection table by splitting the audio data into segments of 
            >>> #uniform length. The length is set to 10.0 sec and the step size to 5.0 sec.
            >>> sel = select_by_segmenting(files=files, length=10.0, annotations=annot, step=5.0) 
            >>> #Inspect the selection table
            >>> print(sel[0].round(2))
                              start   end
            filename  sel_id             
            file1.wav 0         0.0  10.0
                      1         5.0  15.0
            file2.wav 0         0.0  10.0
                      1         5.0  15.0
                      2        10.0  20.0
            file3.wav 0         0.0  10.0
                      1         5.0  15.0
                      2        10.0  20.0
            >>> #Inspect the annotations
            >>> print(sel[1].round(2))
                                       start   end  label
            filename  sel_id annot_id                    
            file1.wav 0      0           7.0   8.1      2
                             1           8.5  12.5      1
                      1      0           2.0   3.1      2
                             1           3.5   7.5      1
                             2           8.1   9.0      2
                      2      1          -1.5   2.5      1
                             2           3.1   4.0      2
            file2.wav 0      0           2.2   3.1      2
                             1           5.8   6.8      2
                             2           9.0  13.0      1
                      1      1           0.8   1.8      2
                             2           4.0   8.0      1
                      2      2          -1.0   3.0      1
    """
    if step is None:
        step = length

    # check that the annotation table has expected format
    if annotations is not None:
        assert is_standardized(annotations, has_time=True), 'Annotation table appears not to have the expected structure.'
        
        annotations = annotations[annotations.label != -1] #discard annotations with label -1

    # create selections table by segmenting
    sel = segment_files(files, length=length, step=step, pad=pad)

    # max number of segments
    num_segs = sel.index.get_level_values(1).max() + 1

    # create annotation table by segmenting
    if annotations is not None:
        annot = segment_annotations(annotations, num=num_segs, length=length, step=step)

        # get the indices of those selections that have annotations associated with them
        indices = list(set([(a, b) for a, b, c in annot.index.tolist()]))

        if keep_only_empty: 
            sel = sel.loc[~sel.index.isin(indices)].sort_index()
            sel['label'] = label_empty
            return sel

        if discard_empty: 
            sel = sel.loc[indices].sort_index()

        if avoid_label is not None:
            if isinstance(avoid_label, int): avoid_label = [avoid_label]
            annot_avoid = annot[annot.label.isin(avoid_label)]
            indices_avoid = list(set([(a, b) for a, b, c in annot_avoid.index.tolist()]))
            sel = sel.loc[~sel.index.isin(indices_avoid)].sort_index()

        return sel, annot
            
    else:
        return sel

def segment_files(table, length, step=None, pad=True):
    """ Generate a selection table by stepping across the audio files, using a fixed 
        step size (step) and fixed selection window size (length). 

        Args:
            table: pandas DataFrame
                File duration table.
            length: float
                Selection length in seconds.
            step: float
                Selection step size in seconds. If None, the step size is set 
                equal to the selection length.
            pad: bool
                If True (default), the last selection window is allowed to extend 
                beyond the endpoint of the audio file.

        Returns:
            df: pandas DataFrame
                Selection table
    """
    if step is None:
        step = length

    # compute number of segments for each file
    table['num'] = (table['duration'] - length) / step + 1
    table.num = table.num.apply(lambda x: np.maximum(x, 0))
    if pad: 
        table.num = table.num.apply(np.ceil).astype(int)
    else:
        table.num = table.num.apply(np.floor).astype(int)

    df = table.loc[table.index.repeat(table.num)]
    df.set_index(keys=['filename'], inplace=True, append=True)
    df = df.swaplevel()
    df = df.sort_index()
    df.index = pd.MultiIndex.from_arrays(
        [df.index.get_level_values(0), df.groupby(level=0).cumcount()],
        names=['filename', 'sel_id'])

    df['start'] = df.index.get_level_values(1) * step
    df['end'] = df['start'] + length
    df.drop(columns=['num','duration'], inplace=True)

    return df

def segment_annotations(table, num, length, step=None, compute_overlap=False):
    """ Generate a segmented annotation table by stepping across the audio files, using a fixed 
        step size (step) and fixed selection window size (length). 
        
        Args:
            table: pandas DataFrame
                Annotation table.
            num: int
                Number of segments
            length: float
                Selection length in seconds.
            step: float
                Selection step size in seconds. If None, the step size is set 
                equal to the selection length.
            compute_overlap: bool
                If True, the fractional overlap between the selection window and the annotation 
                will be computed and added as an extra column in the output table. Default is False.

        Returns:
            df: pandas DataFrame
                Annotations table
    """
    if step is None:
        step = length

    segs = []
    for n in range(num):
        # select annotations that overlap with segment
        t1 = n * step
        t2 = t1 + length
        a = table[(table.start < t2) & (table.end > t1)].copy()
        if len(a) > 0:
            # shift and crop annotations
            if compute_overlap:
                a['overlap'] = a.apply(lambda r: fractional_overlap(a=(t1,t2), b=(r.start, r.end)), axis=1)

            a['start'] -= t1 
            a['end'] -= t1 
            a['sel_id'] = n #map to segment
            segs.append(a)

    df = pd.concat(segs)
    df.set_index(keys=['sel_id'], inplace=True, append=True)
    df = df.swaplevel()
    df = df.sort_index()
    return df

def query(selections, annotations=None, filename=None, label=None, start=None, end=None):
    """ Query selection table for selections from certain audio files 
        and/or with certain labels.

        Args:
            selections: pandas DataFrame
                Selections table
            annotations: pandas DataFrame
                Annotations table. Optional.
            filename: str or list(str)
                Filename(s)
            label: int or list(int)
                Label(s)
            start: float
                Earliest end time in seconds
            end: float
                Latest start time in seconds

        Returns:
            : pandas DataFrame or tuple(pandas DataFrame, pandas DataFrame)
            Selection table, accompanied by an annotation table if an input 
            annotation table is provided.
    """
    if annotations is None:
        return query_labeled(selections, filename, label, start, end)
    else:
        return query_annotated(selections, annotations, filename, label, start, end)

def query_labeled(table, filename=None, label=None, start=None, end=None):
    """ Query selection table for selections from certain audio files 
        and/or with certain labels.

        Args:
            table: pandas DataFrame
                Annotations table or Selections table with a 'label' column.
            filename: str or list(str)
                Filename(s)
            label: int or list(int)
                Label(s)
            start: float
                Earliest end time in seconds
            end: float
                Latest start time in seconds

        Returns:
            df: pandas DataFrame
            Selection table
    """
    df = table
    if filename is not None:
        if isinstance(filename, str): filename = [filename]

        filename = [f for f in filename if f in df.index]

        if len(filename) == 0: 
            return df.iloc[0:0]
        else: 
            df = df.loc[filename]

    if label is not None:
        if not isinstance(label, list):
            label = [label]

        df = df[df.label.isin(label)]

    if start is not None:
        df = df[df.end > start]

    if end is not None:
        df = df[df.start < end]

    return df

def query_annotated(selections, annotations, filename=None, label=None, start=None, end=None):
    """ Query selection table for selections from certain audio files 
        and/or with certain labels.

        Args:
            selections: pandas DataFrame
                Selections table.
            annotations: pandas DataFrame
                Annotations table.
            filename: str or list(str)
                Filename(s)
            label: int or list(int)
                Label(s)
            start: float
                Earliest end time in seconds
            end: float
                Latest start time in seconds

        Returns:
            df1,df2: tuple(pandas DataFrame, pandas DataFrame)
                Selection table and annotation table
    """
    df1 = selections
    df2 = annotations

    df1 = query_labeled(df1, filename=filename, start=start, end=end)
    df2 = query_labeled(df2, filename=filename, label=label, start=start, end=end)

    indices = list(set([x[:-1] for x in df2.index.tolist()]))
    df1 = df1.loc[indices].sort_index()

    return df1, df2

def aggregate_duration(table, label=None):
    """ Compute the aggregate duration of the annotations.

        Overlapping segments are only counted once.

        Args:
            table: pandas DataFrame
                Annotations table or Selections table
            label: int or list(int)
                Label(s). Optional

        Returns:
            agg_dur: float
                Aggregate duration in seconds
    """
    df = query_labeled(table=table, label=label)
    df.sort_index(axis='index', level=[0, 1], inplace=True)

    agg_dur = 0
    
    for index, row in df.iterrows():
        filename = index[0]
        dur = row['end'] - row['start']

        if agg_dur == 0 or filename != filename_prev:
            agg_dur += dur
            end_prev = row['end']
        
        else:
            if row['start'] >= end_prev:
                agg_dur += dur
                end_prev = row['end']

            else:
                extend = max(0, row['end'] - end_prev)                
                agg_dur += extend
                end_prev += extend

        filename_prev = filename

    return agg_dur



from ketos.data_handling.parsing import parse_parameter
from ketos.audio.spectrogram import MagSpectrogram, PowerSpectrogram, MelSpectrogram, CQTSpectrogram
from ketos.audio.waveform import Waveform
from ketos.audio.gammatone import GammatoneFilterBank, AuralFeatures
from pathlib import Path
from zipfile import ZipFile
import warnings
import importlib
import sys
import zipimport
import json

audio_representation_names_in_recipe = {'Waveform':Waveform,
                                        'MagSpectrogram':MagSpectrogram, 
                                        'PowerSpectrogram':PowerSpectrogram,
                                        'MelSpectrogram':MelSpectrogram,
                                        'CQTSpectrogram':CQTSpectrogram,
                                        'AuralFeatures': AuralFeatures,
                                        'GammatoneFilterBank': GammatoneFilterBank}

def load_audio_representation_from_file(model_file):
    """ Load the audio representation from a ketos (.kt) model file.
        
        Args:
            model_file:str
                Path to the ketos(.kt) file
                
        Raises:
            ValueError: If the audio representation configuration does not contain a valid value for the 'type' field.

        Returns:
            audio_repr: Returns a dictionary with the loaded audio representation.

    """
    representation_names = []
    audio_representations = [] # There can be multiple representations

    # First we need to get the recipe file to know which audio representation we are working with.
    path = Path(model_file)
    with ZipFile(path, 'r') as archive: # we dont need to extract from the zipfile, we can just load directly into memory
        audio_recipe_file = archive.open('audio_repr.json')
        json_content = json.load(audio_recipe_file)
        # the representations are stored in the files as: {rep: representation, rep2: representation_2}, 
        for representation_name, representation in json_content.items():
            representation_names.append(representation_name)
            audio_representations.append(representation)
    
    # The script prioritizes custom audio representations over the default ones from ketos. So if the user defined a custom representation with the same name as a default one, it will prioritize the custom
    # if there is no custom representation it will attempt to load a default one
    for idx, representation in enumerate(audio_representations):
        try:
            # See docs https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
            module_name = "audio_representation" #The file name will always be audio_representation because this is how ketos saves it
            module_path = zipimport.zipimporter(path / "custom") # the folder name within the zip archive will always be custom
            spec = module_path.find_spec(module_name)

            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module # adding the module to the list of modules in sys
            spec.loader.exec_module(module) # actually load the module, this is equivalent to importing
            # Now we load the class
            audio_representation_class = getattr(module, representation['type'])
            audio_representations[idx]['type'] = audio_representation_class
            # Warn the user about using a custom input transform function
            warnings.warn("Ketos detected a custom audio representation and will be using that.", UserWarning)
        except AttributeError:
            # fallback to the dictionary
            try:
                # load the class from the default audio representations in ketos
                audio_representations[idx]['type'] = audio_representation_names_in_recipe[representation['type']]
            except KeyError as e:
                e.__suppress_context__ = True # we are supressing the first attribute error as it is not useful
                raise ValueError(f"The audio representation '{representation['type']}' is not a valid audio representation")
        
        # We now need to parse the string values in the representation to their proper units. 
        # ToDo: As we discuseed previously, we should use default units os htat we dont have to make this complicated parsing.

        for key,value in representation.items(): 
            audio_representations[idx][key] = parse_parameter(name=key, value=value)
    audio_representation_dicts = []
    for name, representation in zip(representation_names, audio_representations):
        audio_representation_dicts.append({name: representation})
    return audio_representation_dicts